"""
Absfuyu: Zipped
---------------
Zipping stuff
(deprecated, use absfuyu.util.path.Directory)

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["Zipper"]


# Library
# ---------------------------------------------------------------------------
import shutil
import zipfile
from pathlib import Path

from absfuyu.core import BaseClass, deprecated, versionadded
from absfuyu.logger import logger


# Class
# ---------------------------------------------------------------------------
@deprecated("5.1.0", reason="Use ``absfuyu.util.path.Directory`` instead")
class Zipper(BaseClass):
    """Zip file or folder"""

    def __init__(self, path_to_zip: str | Path, name: str | None = None) -> None:
        """
        path_to_zip: source location
        name: zipped file name
        """
        self.source_path = Path(path_to_zip)
        if name is None:
            self.name = self.source_path.name + ".zip"
        else:
            self.name = name
        self.destination = self.source_path.parent.joinpath(self.name)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def zip_stuff(self, delete_after_zip: bool = False) -> None:
        """
        Zip file/folder

        :param delete_after_zip: Delete source after zip (Default: ``False``)
        :type delete_after_zip: bool
        """

        # Zip
        logger.debug("Zipping...")
        if self.source_path.is_dir():  # zip entire folder
            try:
                with zipfile.ZipFile(self.destination, "w", zipfile.ZIP_DEFLATED) as f:
                    for file in self.source_path.rglob("*"):
                        f.write(file, file.relative_to(self.source_path))
            except Exception:
                logger.error("Zip failed!")
                # shutil.make_archive(zip_file, format="zip", root_dir=zip_path) # Method 2
        else:  # zip a file
            # Implement later
            pass

        # Delete folder
        if delete_after_zip:
            try:
                logger.debug("Deleting unused folder...")
                shutil.rmtree(self.source_path)
                logger.debug("Files deleted")
            except OSError as e:
                logger.error(f"Error: {e.filename} - {e.strerror}.")

    @versionadded("4.0.0")
    def unzip(self):
        """
        Unzip every archive files in directory
        """
        _valid = [".zip", ".cbz"]
        for x in self.source_path.glob("*"):
            if x.suffix.lower() in _valid:
                logger.debug(f"Unzipping {x.name}...")
                if x.suffix.lower() == ".cbz":
                    temp = x.rename(x.with_suffix(".zip"))
                    shutil.unpack_archive(temp, temp.parent.joinpath(temp.stem))
                else:
                    shutil.unpack_archive(x, x.parent.joinpath(x.stem))



"""
ABSFUYU
-------
A small collection of code

LINKS
-----
- [Home page](https://pypi.org/project/absfuyu/)
- [Documentation](https://absolutewinter.github.io/absfuyu/)

USAGE
-----

Normal import:
>>> import absfuyu
>>> help(absfuyu)

Using in cmd:
``$ fuyu --help``
"""

__title__ = "absfuyu"
__author__ = "AbsoluteWinter"
__license__ = "MIT License"
__version__ = "5.15.1"
__all__ = [
    "core",
    "config",
    "dxt",
    "extra",
    "logger",
    "fun",
    "game",
    "general",
    "pkg_data",
    "sort",
    "tools",
    "util",
    "version",
]



"""
ABSFUYU
-------
COMMAND LINE INTERFACE

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

__all__ = ["cli"]

import platform

import click
import colorama

from absfuyu import __title__, __version__
from absfuyu.cli.color import COLOR
from absfuyu.cli.config_group import config_group
from absfuyu.cli.do_group import do_group
from absfuyu.cli.game_group import game_group
from absfuyu.cli.tool_group import tool_group

# Color stuff
colorama.init(autoreset=True)


@click.command()
def version() -> None:
    """Show current version"""
    ver_msg = f"{__title__} v{__version__}"
    click.echo(
        f"{COLOR['green']}{ver_msg}{COLOR['reset']}\n"
        f"- os/type: {platform.system().lower()}\n"
        f"- os/kernel: {platform.version()}\n"
        f"- os/arch: {platform.machine().lower()}\n"
        f"- python version: {platform.python_version()}\n"
    )


@click.group(name="cli")
def cli() -> None:
    """
    absfuyu's command line interface
    """
    pass


cli.add_command(config_group)
cli.add_command(do_group)
cli.add_command(game_group)
cli.add_command(tool_group)
cli.add_command(version)



"""
Absfuyu: Image duplicate checker
--------------------------------
Image duplicate checker


Version: 5.15.0
Date updated: 05/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["DirectoryRemoveDuplicateImageMixin"]


# Library
# ---------------------------------------------------------------------------
from dataclasses import dataclass
from enum import StrEnum
from functools import partial, total_ordering
from pathlib import Path
from typing import Callable, Literal, NamedTuple

from absfuyu.core.dummy_func import tqdm as tqdm_base
from absfuyu.tools.checksum import DirectoryRemoveDuplicateMixin, DuplicateSummary

try:
    import imagehash
    from PIL import Image
except ImportError:
    from subprocess import run

    from absfuyu.config import ABSFUYU_CONFIG

    if ABSFUYU_CONFIG._get_setting("auto-install-extra").value:  # type: ignore
        cmd = "python -m pip install -U absfuyu[pic]".split()
        run(cmd)
    else:
        raise SystemExit("This feature is in absfuyu[pic] package")  # noqa: B904

# Setup
# ---------------------------------------------------------------------------
tqdm = partial(tqdm_base, unit_scale=True, dynamic_ncols=True)
SupportedImageFormat = {".jpg", ".jpeg", ".png", ".webp", ".bmp"}


# Class
# ---------------------------------------------------------------------------
class HashMode(StrEnum):
    PERCEPTUAL_HASH = "phash"
    AVERAGE_HASH = "ahash"
    DIFFERENCE_HASH = "dhash"
    WAVELET_HASH = "whash"


class DuplicateImgPair(NamedTuple):
    """
    Duplicate image pair

    Parameters
    ----------
    original : Path
        Original image path

    duplicate : Path
        Duplicate image path

    distant : int
        Similarity between image (0 is exact)
    """

    original: Path
    duplicate: Path
    distant: int


@total_ordering
@dataclass
class ImageInfo:
    """
    Quick image info

    Parameters
    ----------
    path : Path
        Image path

    file_size : int
        File size

    dimension : tuple[int, int]
        Dimension (width, height)
    """

    path: Path
    file_size: int
    dimension: tuple[int, int]

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            raise NotImplementedError("Not implemented")
        return self.dimension == other.dimension and self.file_size == other.file_size

    def __lt__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            raise NotImplementedError("Not implemented")

        # prioritize dimension first, then size
        if self.dimension != other.dimension:
            return self.dimension < other.dimension
        return self.file_size < other.file_size


class DirectoryRemoveDuplicateImageMixin(DirectoryRemoveDuplicateMixin):
    """
    Directory - Remove duplicate image

    - remove_duplicate_images


    Example:
    --------
    >>> DirectoryRemoveDuplicateImageMixin(".").remove_duplicate_images()
    """

    def __init__(self, source_path, create_if_not_exist=False) -> None:
        super().__init__(source_path, create_if_not_exist)

        # Unused yet
        self._duplicate_image_cache = None

    # Hash
    def _get_img_hash_mode(
        self, hash_mode: HashMode = HashMode.PERCEPTUAL_HASH
    ) -> Callable[[Image, int], imagehash.ImageHash]:
        """
        Get image hash mode

        Parameters
        ----------
        hash_mode : HashMode, optional
            Hash mode, by default ``HashMode.PERCEPTUAL_HASH``

        Returns
        -------
        Callable[[Image, int], imagehash.ImageHash]
            Hash function
        """
        if hash_mode == HashMode.AVERAGE_HASH:
            return imagehash.average_hash
        elif hash_mode == HashMode.DIFFERENCE_HASH:
            return imagehash.dhash
        elif hash_mode == HashMode.WAVELET_HASH:
            return imagehash.whash
        else:
            return imagehash.phash

    def _gather_duplicate_image_cache(
        self, recursive: bool = True, threshold: int = 5, hash_mode: HashMode = HashMode.PERCEPTUAL_HASH
    ) -> None:
        """
        Gather duplicate image cache

        Parameters
        ----------
        recursive : bool, optional
            Scan every file in the folder (including child folder), by default ``True``

        threshold : int, optional
            Maximum hamming distance between image hashes to consider them "similar", by default ``5``
            - 0: Exact image
            - [5,10]: Tolerant of light edits

        hash_mode : HashMode, optional
            Hash mode, by default ``HashMode.PERCEPTUAL_HASH``
        """
        valid = [
            x
            for x in self.source_path.glob("**/*" if recursive else "*")
            if x.is_file() and x.suffix.lower() in SupportedImageFormat
        ]
        hash_cache: dict[imagehash.ImageHash, list[Path]] = {}
        duplicates: list[DuplicateImgPair] = []

        # Checksum
        for x in tqdm(valid, desc="Hashing image..."):
            try:
                with Image.open(x) as img:
                    hash_func = self._get_img_hash_mode(hash_mode=hash_mode)
                    hash = hash_func(img)  # perceptual hash

            except Exception as err:
                print(f"ERROR: {x} - {err}")
                continue

            # Compare against all cached hashes
            found = False
            for existing_hash, paths in hash_cache.items():
                distance = hash - existing_hash
                if distance <= threshold:
                    duplicates.append(DuplicateImgPair(paths[0], x, distance))
                    if x not in paths:
                        paths.append(x)
                    found = True
                    break

            if not found:
                hash_cache[hash] = [x]

        # Save to cache
        self._duplicate_cache = DuplicateSummary({k: v for k, v in hash_cache.items() if len(v) > 1})
        self._duplicate_image_cache = duplicates

    # Remove
    def _gather_img_info(self, image_path: Path) -> ImageInfo:
        with Image.open(image_path) as img:
            dim = img.size
        return ImageInfo(image_path, image_path.stat().st_size, dim)

    def _remove_duplicate_image_best(self, dry_run: bool = True, debug: bool = True) -> None:
        """This will take image with large size in dimension and storage"""
        if self._duplicate_cache is None or self._duplicate_image_cache is None:
            raise ValueError("No duplicates found")

        del_list: list[ImageInfo] = []
        for paths in self._duplicate_cache.values():
            # Sort image by dimension then size ascending order then cut the last value
            data = sorted([self._gather_img_info(img) for img in paths])[:-1]
            # Extend to delete list
            del_list.extend(data)

        for i, x in enumerate(del_list, start=1):
            if debug:
                print(f"{i:02}. Deleting {x.path}")
            if not dry_run:
                x.path.unlink(missing_ok=True)

    # Main
    def remove_duplicate_images(
        self,
        dry_run: bool = True,
        recursive: bool = True,
        threshold: int = 5,
        hash_mode: HashMode = HashMode.PERCEPTUAL_HASH,
        keep_mode: Literal["first", "last", "best"] = "best",
        debug: bool = True,
    ) -> None:
        """
        Remove duplicate images in a directory

        Parameters
        ----------
        dry_run : bool, optional
            Simulate only (no files deleted), by default ``True``

        recursive : bool, optional
            Scan every file in the folder (including child folder), by default ``True``

        threshold : int, optional
            Maximum hamming distance between image hashes to consider them "similar", by default ``5``
            - 0: Exact image
            - [5,10]: Tolerant of light edits

        hash_mode : HashMode, optional
            Hash mode, by default ``HashMode.PERCEPTUAL_HASH``

        keep_mode : Literal["first", "last", "best"], optional
            What to keep in duplicate images, by default ``"best"``
            - "first": First item in delete list
            - "last": Last item in delete list
            - "best": Best item (largest dimension and size) in delete list

        debug : bool, optional
            Debug message, by default ``True``
        """
        # Cache
        self._gather_duplicate_image_cache(recursive=recursive, threshold=threshold, hash_mode=hash_mode)

        # Remove
        try:
            if keep_mode in ["first", "last"]:
                summary = self._duplicate_cache
                print(f"Duplicate files: {summary.summary()}")
                summary.remove_duplicates(dry_run=dry_run, keep_first=keep_mode == "first", debug=debug)

            else:  # best mode
                self._remove_duplicate_image_best(dry_run=dry_run, debug=debug)

        except Exception as err:
            pass



"""
Game: Schulte
-------------

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["SchulteTable"]


# Library
# ---------------------------------------------------------------------------
from absfuyu.core.baseclass import BaseClass
from absfuyu.dxt import ListExt
from absfuyu.util.text_table import BoxStyle, get_box_drawing_character


# Function
# ---------------------------------------------------------------------------
def draw_grid(data: list[list[int]], style: BoxStyle = "normal"):
    chars = get_box_drawing_character(style=style)
    rows = len(data)
    cols = len(data[0])

    # find max width for padding
    cell_width = max(len(str(x)) for row in data for x in row) + 2

    def horizontal_border(left: str, middle: str, right: str) -> str:
        return left + (chars.HORIZONTAL * cell_width + middle) * (cols - 1) + chars.HORIZONTAL * cell_width + right

    # top border
    print(horizontal_border(chars.UPPER_LEFT_CORNER, chars.HORIZONTAL_DOWN, chars.UPPER_RIGHT_CORNER))

    for i, row in enumerate(data):
        # content line
        line = chars.VERTICAL + "".join(f"{str(val):^{cell_width}}" + chars.VERTICAL for val in row)
        print(line)

        # middle or bottom border
        if i < rows - 1:
            print(horizontal_border(chars.VERTICAL_RIGHT, chars.CROSS, chars.VERTICAL_LEFT))
        else:
            print(horizontal_border(chars.LOWER_LEFT_CORNER, chars.HORIZONTAL_UP, chars.LOWER_RIGHT_CORNER))


# Class
# ---------------------------------------------------------------------------
class SchulteTable(BaseClass):
    """
    A Schulte Table is a cognitive training tool consisting of a grid filled
    with randomly placed numbers. The task is to find and select all the numbers
    in ascending order as quickly as possible. This exercise helps improve
    visual attention, focus, processing speed, mental flexibility,
    and peripheral vision.
    """

    def __init__(self, size: int = 5) -> None:
        self.size = max(size, 1)

    def make_table(self) -> None:
        data = ListExt(range(1, self.size**2 + 1)).shuffle().split_chunk(self.size)
        draw_grid(data)

    # def play(self):
    #     """GUI"""
    #     from absfuyu.util.gui import CustomTkinterApp

    #     class Schulte(CustomTkinterApp):
    #         def __init__(self, title: str | None = None, size: tuple[int, int] | None = None) -> None:
    #             super().__init__(title=title, size=size)


if __name__ == "__main__":
    t = SchulteTable(5)
    print(t)



"""
Global fixture

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import pytest


@pytest.fixture(scope="session")
def test_fixture_session():
    """This cache the fixture for current test session"""
    return None



"""
Absfuyu: Core
-------------
Bases for other features

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    # color
    "CLITextColor",
    # path
    # "CORE_PATH",
    # class
    "GetClassMembersMixin",
    "BaseClass",
    # wrapper
    "tqdm",
    "unidecode",
    # decorator
    "deprecated",
    "versionadded",
    "versionchanged",
]

__package_feature__ = [
    "beautiful",  # BeautifulOutput
    "docs",  # For (package) hatch's env use only
    "extra",  # DataFrame
    "full",  # All package
    "dev",
]


# Library
# ---------------------------------------------------------------------------
# from importlib.resources import files

# Most used features are imported to core
from absfuyu.core.baseclass import BaseClass, CLITextColor, GetClassMembersMixin
from absfuyu.core.docstring import deprecated, versionadded, versionchanged
from absfuyu.core.dummy_func import tqdm, unidecode

# Path
# ---------------------------------------------------------------------------
# CORE_PATH = files("absfuyu")



"""
Game: Game Stat

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["GameStats"]


# Library
# ---------------------------------------------------------------------------
from dataclasses import dataclass, field
from typing import Literal


# Class
# ---------------------------------------------------------------------------
@dataclass
class GameStats:
    win: int = field(default=0)
    draw: int = field(default=0)
    lose: int = field(default=0)
    win_rate: str = field(init=False)
    _win_rate: float = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self._update_win_rate()

    def _update_win_rate(self) -> None:
        try:
            self._win_rate = self.win / (self.win + self.draw + self.lose)
            self.win_rate = f"{self._win_rate * 100:,.2f}%"
        except ZeroDivisionError:
            self._win_rate = 0
            self.win_rate = "N/A"

    def update_score(self, option: Literal["win", "draw", "lose"]) -> None:
        self.__setattr__(option, self.__getattribute__(option) + 1)
        self._update_win_rate()



"""
Test: Game

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

from typing import Literal

import pytest

from absfuyu.game.game_stat import GameStats


class TestGameStat:
    @pytest.mark.parametrize("score_option", ["win", "draw", "lose"])
    def test_update_score(self, score_option: Literal["win", "draw", "lose"]) -> None:
        instance = GameStats()
        before = getattr(instance, score_option)
        instance.update_score(score_option)
        after = getattr(instance, score_option)
        assert after - before == 1

    @pytest.mark.parametrize("score_option", ["w", "d", "l"])
    def test_update_score_error(
        self, score_option: Literal["win", "draw", "lose"]
    ) -> None:
        instance = GameStats()
        with pytest.raises(AttributeError) as excinfo:
            instance.update_score(score_option)
        assert str(excinfo.value)



"""
Absfuyu: GUI
------------
Custom tkinter GUI

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["CustomTkinterApp"]


# Library
# ---------------------------------------------------------------------------
import tkinter as tk

from absfuyu.pkg_data.logo import AbsfuyuLogo


# Class
# ---------------------------------------------------------------------------
class CustomTkinterApp(tk.Tk):

    def __init__(self, title: str | None = None, size: tuple[int, int] | None = None) -> None:
        """
        Custom Tkinter GUI

        Parameters
        ----------
        title : str | None, optional
            Title of the app, by default None

        size : tuple[int, int] | None, optional
            Size of the app (width, height), by default None
        """
        super().__init__()

        # Set custom icon
        self.iconphoto(True, tk.PhotoImage(data=AbsfuyuLogo.SHORT))

        # Title
        self.title(title)

        # Set size
        self._absfuyu_set_width_height(size)

    # @versionadded("5.10.0")
    def _absfuyu_set_width_height(self, size: tuple[int, int] | None = None) -> None:
        """
        Set width and height for the app.

        Parameters
        ----------
        size : tuple[int, int] | None, optional
            Size of the app (width, height), by default ``None``
            (420 x 250)
        """
        # Set GUI appears in center
        if size is None:
            _width = 420
            _height = 250
        else:
            _width, _height = size

        _width_screen = self.winfo_screenwidth()
        _width_screen_offset = 0.0052  # x offset mutiplier
        _x_offset = int(_width_screen * _width_screen_offset)

        _height_screen = self.winfo_screenheight()
        _height_screen_offset = 0.0926  # y offset mutiplier
        _y_offset = int(_height_screen * _height_screen_offset)

        if _width > _width_screen:
            _width = int(_width_screen * (1 - 0.001))
            # _x_offset = 0

        if _height > (_height_screen - _y_offset):
            _height = int(_height_screen * (1 - 0.001))
            _y_offset = 0

        _x = (_width_screen / 2) - (_width / 2) - _x_offset
        _y = (_height_screen / 2) - (_height / 2) - _y_offset

        self.geometry(f"{_width}x{_height}+{int(_x)}+{int(_y)}")


if __name__ == "__main__":
    app = CustomTkinterApp("absfuyu")
    app.mainloop()



"""
Test: Data extension - DictExt

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import pytest

from absfuyu.dxt import DictAnalyzeResult, DictExt


@pytest.fixture
def dict_example():
    return DictExt({"Line 1": 99, "Line 2": 50})


@pytest.fixture
def dict_example_2():
    return DictExt({"Line 1": 99, "Line 2": "test"})


@pytest.mark.abs_dxt
class TestDictExt:
    """absfuyu.dxt.DictExt"""

    # analyze
    def test_analyze(self, dict_example: DictExt) -> None:
        # assert example.analyze() == {'max_value': 99, 'min_value': 50, 'max': [('Line 1', 99)], 'min': [('Line 2', 50)]}
        assert dict_example.analyze() == DictAnalyzeResult(
            99, 50, [("Line 1", 99)], [("Line 2", 50)]
        )

    def test_analyze_error(self, dict_example_2: DictExt) -> None:
        """When values are not int or float"""
        with pytest.raises(ValueError) as excinfo:
            dict_example_2.analyze()
        assert str(excinfo.value)

    # swap
    def test_swap(self, dict_example: DictExt) -> None:
        assert dict_example.swap_items() == {99: "Line 1", 50: "Line 2"}

    # apply
    def test_apply(self, dict_example: DictExt) -> None:
        """Values"""
        assert dict_example.apply(str) == {"Line 1": "99", "Line 2": "50"}

    def test_apply_2(self) -> None:
        """Keys"""
        assert DictExt({1: 1}).apply(str, apply_to_value=False) == {"1": 1}

    # aggregate
    def test_aggregate(self, dict_example: DictExt) -> None:
        agg = {"Line 1": 1, "Line 3": 1}
        new_dict = dict_example.aggregate(agg)
        assert new_dict == {"Line 1": 100, "Line 2": 50, "Line 3": 1}

    def test_aggregate_2(self, dict_example: DictExt) -> None:
        """Empty dict"""
        new_dict = DictExt().aggregate(dict_example)
        assert new_dict == dict_example

    def test_aggregate_3(self, dict_example: DictExt) -> None:
        """Different type of data"""
        agg = {"Line 1": "1", "Line 3": 1}
        new_dict = dict_example.aggregate(agg)
        assert new_dict == {"Line 1": [99, "1"], "Line 2": 50, "Line 3": 1}



"""
Absfuyu: Version
----------------
Package versioning module

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    # Options
    "ReleaseOption",
    "ReleaseLevel",
    # Class
    "Version",
    "Bumper",
    "PkgVersion",
]


# Library
# ---------------------------------------------------------------------------
import json
import re
import subprocess
from enum import StrEnum
from typing import Self, TypedDict
from urllib.error import URLError
from urllib.request import Request, urlopen

from absfuyu.core import BaseClass
from absfuyu.logger import logger


# Class
# ---------------------------------------------------------------------------
class ReleaseOption(StrEnum):
    """
    ``MAJOR``, ``MINOR``, ``PATCH``
    """

    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"

    @classmethod
    def all_option(cls) -> list[str]:
        """Return a list of release options"""
        return [cls.MAJOR.value, cls.MINOR.value, cls.PATCH.value]


class ReleaseLevel(StrEnum):
    """
    ``FINAL``, ``DEV``, ``RC``
    """

    FINAL = "final"
    DEV = "dev"
    RC = "rc"  # Release candidate

    @classmethod
    def all_level(cls) -> list[str]:
        """Return a list of release levels"""
        return [cls.FINAL.value, cls.DEV.value, cls.RC.value]


class VersionDictFormat(TypedDict):
    """
    Format for the ``version`` section in ``config``

    :param major: Major changes
    :param minor: Minor changes
    :param patch: Patches and fixes
    :param release_level: Release level
    :param serial: Release serial
    """

    major: int
    minor: int
    patch: int
    release_level: str
    serial: int


class Version(BaseClass):
    """Version"""

    def __init__(
        self,
        major: int | str,
        minor: int | str,
        patch: int | str,
        release_level: str = ReleaseLevel.FINAL,
        serial: int | str = 0,
    ) -> None:
        """
        Create ``Version`` instance

        Parameters
        ----------
        major : int | str
            Major change

        minor : int | str
            Minor change

        patch : int | str
            Patch

        release_level : str, optional
            Release level: ``final`` | ``rc`` | ``dev``, by default ``ReleaseLevel.FINAL``

        serial : int | str, optional
            Serial for release level ``rc`` | ``dev``, by default ``0``
        """
        self.major: int = major if isinstance(major, int) else int(major)
        self.minor: int = minor if isinstance(minor, int) else int(minor)
        self.patch: int = patch if isinstance(patch, int) else int(patch)
        self.release_level: str = release_level
        self.serial: int = serial if isinstance(serial, int) else int(serial)

    def __str__(self) -> str:
        return self.version

    # def __repr__(self) -> str:
    #     cls_name = self.__class__.__name__
    #     if self.release_level.startswith(ReleaseLevel.FINAL):
    #         return f"{cls_name}(major={self.major}, minor={self.minor}, patch={self.patch})"
    #     else:
    #         return (
    #             f"{cls_name}("
    #             f"major={self.major}, minor={self.minor}, patch={self.patch}, "
    #             f"release_level={self.release_level}, serial={self.serial})"
    #         )

    def __format__(self, format_spec: str) -> str:
        """
        Change format of an object.
        Avaiable option: ``full``

        Usage
        -----
        >>> print(f"{<object>:<format_spec>}")
        >>> print(<object>.__format__(<format_spec>))
        >>> print(format(<object>, <format_spec>))

        Example:
        --------
        >>> test = Version(1, 0, 0)
        >>> print(f"{test:full}")
        1.0.0.final0
        """
        # Logic
        if format_spec.lower().startswith("full"):
            return f"{self.major}.{self.minor}.{self.patch}.{self.release_level}{self.serial}"

        # Else
        return self.__str__()

    @property
    def version(self) -> str:
        """
        Return version string

        Example:
        --------
        >>> test = Version(1, 0, 0)
        >>> test.version
        1.0.0
        >>> str(test)  # test.__str__()
        1.0.0

        >>> test_serial = Version(1, 0, 0, "dev", 1)
        >>> test_serial.version
        1.0.0.dev1
        """
        if self.release_level.startswith(ReleaseLevel.FINAL):
            return f"{self.major}.{self.minor}.{self.patch}"
        else:
            return f"{self.major}.{self.minor}.{self.patch}.{self.release_level}{self.serial}"

    @classmethod
    def from_tuple(
        cls, iterable: tuple[int, int, int] | tuple[int, int, int, str, int]
    ) -> Self:
        """
        Convert to ``Version`` from a ``tuple``

        Parameters
        ----------
        iterable : tuple[int, int, int] | tuple[int, int, int, str, int]
            Version tuple in correct format

        Returns
        -------
        Version
            Version

        Raises
        ------
        ValueError
            Wrong tuple format


        Example:
        --------
        >>> test = Version.from_tuple((1, 0, 0))
        >>> test.version
        1.0.0
        """
        if len(iterable) == 5:
            return cls(
                iterable[0], iterable[1], iterable[2], iterable[3], iterable[4]
            )  # Full
        elif len(iterable) == 3:
            return cls(iterable[0], iterable[1], iterable[2])  # major.minor.patch only
        else:
            raise ValueError("iterable must have len of 5 or 3")

    @classmethod
    def from_str(cls, version_string: str) -> Self:
        """
        Convert to ``Version`` from a ``str``

        Parameters
        ----------
        version_string : str
            | Version str in correct format
            | ``<major>.<minor>.<patch>``
            | ``<major>.<minor>.<patch>.<release level><serial>``

        Returns
        -------
        Version
            Version

        Raises
        ------
        ValueError
            Wrong version_string format


        Example:
        --------
        >>> test = Version.from_str("1.0.0")
        >>> test.version
        1.0.0
        """
        short_ver_pattern = re.compile(r"\b(\d)+\.(\d+)\.(\d+)\b")
        long_ver_pattern = re.compile(r"\b(\d)+\.(\d+)\.(\d+)\.(dev|rc|final)(\d+)\b")
        ver = version_string.lower().strip()

        long_ver = re.search(long_ver_pattern, ver)
        if long_ver:
            return cls.from_tuple(long_ver.groups())  # type: ignore

        short_ver = re.search(short_ver_pattern, ver)
        if short_ver:
            return cls.from_tuple(short_ver.groups())  # type: ignore

        raise ValueError("Wrong version_string format")

    def to_dict(self) -> VersionDictFormat:
        """
        Convert ``Version`` into ``dict``

        Returns
        -------
        VersionDictFormat
            Version dict


        Example:
        --------
        >>> test = Version(1, 0, 0)
        >>> test.to_dict()
        {
            "major": 1,
            "minor": 0,
            "patch": 0,
            "release_level": "final",
            "serial": 0
        }
        """
        out: VersionDictFormat = {
            "major": self.major,
            "minor": self.minor,
            "patch": self.patch,
            "release_level": self.release_level,
            "serial": self.serial,
        }
        return out


class Bumper(Version):
    """Version bumper"""

    def _bump_ver(self, release_option: str) -> None:
        """
        Bumping major, minor, patch
        """

        if release_option.startswith(ReleaseOption.MAJOR):
            self.major += 1
            self.minor = 0
            self.patch = 0
        elif release_option.startswith(ReleaseOption.MINOR):
            self.minor += 1
            self.patch = 0
        else:
            self.patch += 1

    def bump(
        self, *, option: str = ReleaseOption.PATCH, channel: str = ReleaseLevel.FINAL
    ) -> None:
        """
        Bump current version (internally)

        Parameters
        ----------
        option : str
            Release option
            (Default: ``"patch"``)

        channel : str
            Release channel
            (Default: ``"final"``)


        Example:
        --------
        >>> test = Bumper(1, 0, 0)
        >>> test.version
        1.0.0
        >>> test.bump()
        >>> test.version
        1.0.1
        """
        # Check conditions - use default values if fail
        if option not in ReleaseOption.all_option():
            logger.warning(f"Available option: {ReleaseOption.all_option()}")
            option = ReleaseOption.PATCH
        if channel not in ReleaseLevel.all_level():
            logger.warning(f"Available level: {ReleaseLevel.all_level()}")
            channel = ReleaseLevel.FINAL
        logger.debug(f"Target: {option} {channel}")

        # Bump ver
        if channel.startswith(ReleaseLevel.FINAL):  # Final release level
            if self.release_level in [
                ReleaseLevel.RC,
                ReleaseLevel.DEV,
            ]:  # current release channel is dev or rc
                self.release_level = ReleaseLevel.FINAL
                self.serial = 0
            else:
                self.serial = 0  # final channel does not need serial
                self._bump_ver(option)

        elif channel.startswith(ReleaseLevel.RC):  # release candidate release level
            if self.release_level.startswith(
                ReleaseLevel.DEV
            ):  # current release channel is dev
                self.release_level = ReleaseLevel.RC
                self.serial = 0  # reset serial
            elif channel == self.release_level:  # current release channel is rc
                self.serial += 1
            else:  # current release channel is final
                self.release_level = channel
                self.serial = 0  # reset serial
                self._bump_ver(option)

        else:  # dev release level
            if channel == self.release_level:  # current release channel is dev
                self.serial += 1
            else:  # current release channel is final or rc
                self.release_level = channel
                self.serial = 0
                self._bump_ver(option)


class PkgVersion:
    """
    Package Version
    """

    def __init__(self, package_name: str) -> None:
        self.package_name = package_name

    # Check for update
    @staticmethod
    def _fetch_data_from_server(link: str):
        """Fetch data from API"""
        req = Request(link)
        try:
            response = urlopen(req)
            # return response
        except URLError as e:
            if hasattr(e, "reason"):
                logger.error("Failed to reach server.")
                logger.error("Reason: ", e.reason)
            elif hasattr(e, "code"):
                logger.error("The server couldn't fulfill the request.")
                logger.error("Error code: ", e.code)
        except Exception:
            logger.error("Fetch failed!")
        else:
            return response.read().decode()

    def _get_latest_version_legacy(self) -> str:
        """
        Load data from PyPI's RSS -- OLD
        """
        rss = f"https://pypi.org/rss/project/{self.package_name}/releases.xml"
        xml_file: str = self._fetch_data_from_server(rss)
        ver = xml_file[
            xml_file.find("<item>") : xml_file.find(
                "</item>"
            )  # noqa: E203
        ]  # First item
        version = ver[
            ver.find("<title>") + len("<title>") : ver.find(
                "</title>"
            )  # noqa: E203
        ]
        return version

    def _load_data_from_json(self, json_link: str) -> dict:
        """
        Load data from api then convert to json
        """
        json_file: str = self._fetch_data_from_server(json_link)
        return json.loads(json_file)  # type: ignore

    def _get_latest_version(self) -> str:
        """
        Get latest version from PyPI's API
        """
        link = f"https://pypi.org/pypi/{self.package_name}/json"
        ver: str = self._load_data_from_json(link)["info"]["version"]
        logger.debug(f"Latest: {ver}")
        return ver

    def _get_update(self):
        """
        Run pip upgrade command
        """
        cmd = f"pip install -U {self.package_name}".split()
        return subprocess.run(cmd)

    def check_for_update(
        self,
        *,
        force_update: bool = False,
    ) -> None:
        """
        Check for latest update

        :param force_update: Auto update the package when run (Default: ``False``)
        :type force_update: bool
        """

        try:
            latest = self._get_latest_version()
        except Exception:
            latest = self._get_latest_version_legacy()

        try:
            import importlib

            _pk = importlib.__import__(self.package_name)
            current: str = _pk.__version__
        except Exception:
            current = ""

        logger.debug(f"Current: {current} | Lastest: {latest}")

        if current == latest:
            print(f"You are using the latest version ({latest})")
        else:
            if force_update:
                print(f"Newer version ({latest}) available. Upgrading...")
                try:
                    self._get_update()
                except Exception:
                    print(
                        f"""
                    Unable to perform update.
                    Please update manually with:
                    pip install -U {self.package_name}=={latest}
                    """
                    )
            else:
                print(
                    f"Newer version ({latest}) available. Upgrade with:\npip install -U {self.package_name}=={latest}"
                )



"""
Absfuyu: Inspector
------------------
Inspector

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["Inspector", "inspect_all"]


# Library
# ---------------------------------------------------------------------------
import inspect as _inspect
import os
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from textwrap import TextWrapper
from textwrap import shorten as text_shorten
from typing import Any, Literal, Protocol, get_overloads, overload

from absfuyu.core.baseclass import (
    AutoREPRMixin,
    BaseDataclass,
    ClassMembers,
    ClassMembersResult,
    GetClassMembersMixin,
)
from absfuyu.dxt.listext import ListExt
from absfuyu.typings import P, R
from absfuyu.util.text_table import BoxStyle, OneColumnTableMaker


# TODO: rewrite with each class for docs, method, property, attr, param, title
# Dataclass
# ---------------------------------------------------------------------------
@dataclass
class BaseDCInspect(BaseDataclass):
    def _make_repr(self) -> str | None:
        fields = self._get_fields()
        if 0 < len(fields) < 2:
            return repr(getattr(self, fields[0]))
        return None

    def _long_list_terminal_size(
        self, long_list: list[str], width: int = 80, /
    ) -> list[str]:
        ll = ListExt(long_list).wrap_to_column(width, margin=4, transpose=True)
        return list(ll)


@dataclass
class _TitleSignature:
    """
    Inspector:
    Object's title and signature
    """

    title: str
    signature: list[str]

    def make_output(self):
        pass


@dataclass
class _Docstring(BaseDCInspect):
    """
    Inspector:
    Object's docstring
    """

    docs: str

    def __repr__(self) -> str:
        r = self._make_repr()
        if r is not None:
            return r
        return super().__repr__()

    def _get_first_paragraph(self) -> str:
        # Get docs and get first paragraph
        doc_lines = []
        for line in self.docs.splitlines():
            if len(line) < 1:
                break
            doc_lines.append(line.strip())
        return " ".join(doc_lines)

    def make_output(self) -> list[str]:
        if len(self._get_first_paragraph()) > 0:
            return ["Docstring:", self._get_first_paragraph()]
        return [""]


@dataclass
class _MRO(BaseDCInspect):
    """
    Inspector:
    Object's MRO, bases
    """

    mro: tuple[type, ...]

    def __repr__(self) -> str:
        r = self._make_repr()
        if r is not None:
            return r
        return super().__repr__()

    def _make_output(self) -> list[str]:
        out = [
            f"- {i:02}. {x.__module__}.{x.__name__}"
            for i, x in enumerate(self.mro[1:], start=1)
        ]
        # ListExt.wrap_to_column
        return out

    def make_output(self, *, width: int = 80) -> list[str]:
        out = self._make_output()
        if len(out) > 0:
            out_ = self._long_list_terminal_size(out, width)
            return ["", f"Bases (Len: {len(out)}):", *out_]
        return [""]


@dataclass
class _Member(BaseDCInspect):
    """
    Inspector:
    Object's member
    """

    member: ClassMembers

    def __repr__(self) -> str:
        r = self._make_repr()
        if r is not None:
            return r
        return super().__repr__()

    def make_output(
        self,
        *,
        width: int = 80,
        obj,
        include_method: bool = True,
        include_property: bool = True,
    ) -> list[str]:
        mems = self.member.pack().sort()
        body: list[str] = []

        if include_method:
            ml = [text_shorten(f"- {x}", width - 4) for x in mems.methods]
            if len(ml) > 0:
                head = ["", f"Methods (Len: {len(ml)}):"]
                head.extend(self._long_list_terminal_size(ml, width))
                body.extend(head)

        if include_property:
            pl = [
                text_shorten(f"- {x} = {getattr(obj, x, None)}", width - 4)
                for x in mems.properties
            ]
            if len(pl) > 0:
                head = ["", f"Properties (Len: {len(pl)}):"]
                head.extend(self._long_list_terminal_size(pl, width))
                body.extend(head)

        if len(body) > 0:
            return body
        return [""]


class InspectComponents(Protocol):
    """Supports make_output() -> list[str]"""

    @overload
    def make_output(self) -> list[str]: ...
    @overload
    def make_output(self, *, width: int = ...) -> list[str]: ...
    def make_output(self, *args, **kwargs) -> list[str]: ...


# Class
# ---------------------------------------------------------------------------
class Inspector(AutoREPRMixin):
    """
    Inspect an object.
    By default shows object's docstring and attribute (if any).

    Parameters
    ----------
    obj : Any
        Object to inspect

    line_length: int | None
        Number of cols in inspect output (Split line every line_length).
        Set to ``None`` to use ``os.get_terminal_size()``, by default ``None``

    include_docs : bool, optional
        Include docstring, by default ``True``

    include_mro : bool, optional
        Include class bases (__mro__), by default ``False``

    include_method : bool, optional
        Include object's methods (if any), by default ``False``

    include_property : bool, optional
        Include object's properties (if any), by default ``False``

    include_attribute : bool, optional
        Include object's attributes (if any), by default ``True``

    include_private : bool, optional
        Include object's private attributes, by default ``False``

    include_all : bool, optional
        Include all infomation availble, by default ``False``

    max_textwrap_lines : int, optional
        Maximum lines for the output's header (class, signature, repr).
        Must be >= 1, by default ``8``

    style : BoxStyle, optional
        Style for the table, by default ``"normal"``


    Example:
    --------
    >>> print(Inspector(<object>, **kwargs))
    """

    @overload
    def __init__(self, obj: Any) -> None: ...

    @overload
    def __init__(self, obj: Any, *, include_all: Literal[True] = ...) -> None: ...

    @overload
    def __init__(
        self,
        obj: Any,
        *,
        line_length: int | None = None,
        include_docs: bool = True,
        include_mro: bool = False,
        include_method: bool = False,
        include_property: bool = False,
        include_attribute: bool = True,
        include_private: bool = False,
        max_textwrap_lines: int = 8,
        style: BoxStyle = "normal",
    ) -> None: ...

    def __init__(
        self,
        obj: Any,
        *,
        # Line length
        line_length: int | None = None,
        line_length_offset: int = 0,  # line_length += line_length_offset (when line_length=None)
        max_textwrap_lines: int = 8,
        # Include
        include_docs: bool = True,
        include_mro: bool = False,
        include_method: bool = False,
        include_property: bool = False,
        include_attribute: bool = True,
        include_dunder: bool = False,
        include_private: bool = False,
        include_all: bool = False,
        # Style
        style: BoxStyle = "normal",
    ) -> None:
        """
        Inspect an object.
        By default shows object's docstring and attribute (if any).

        Parameters
        ----------
        obj : Any
            Object to inspect

        line_length: int | None
            Number of cols in inspect output (Split line every line_length).
            Set to ``None`` to use ``os.get_terminal_size()``, by default ``None``

        include_docs : bool, optional
            Include docstring, by default ``True``

        include_mro : bool, optional
            Include class bases (__mro__), by default ``False``

        include_method : bool, optional
            Include object's methods (if any), by default ``False``

        include_property : bool, optional
            Include object's properties (if any), by default ``False``

        include_attribute : bool, optional
            Include object's attributes (if any), by default ``True``

        include_private : bool, optional
            Include object's private attributes, by default ``False``

        include_all : bool, optional
            Include all infomation availble, by default ``False``

        max_textwrap_lines : int, optional
            Maximum lines for the output's header (class, signature, repr).
            Must be >= 1, by default ``8``

        style : BoxStyle | Literal["normal", "bold", "dashed", "double", "rounded", ...], optional
            Style for the table, by default ``"normal"``


        Example:
        --------
        >>> print(Inspector(<object>, **kwargs))
        """
        self.obj = obj
        self.include_docs = include_docs
        self.include_mro = include_mro
        self.include_method = include_method
        self.include_property = include_property
        self.include_attribute = include_attribute
        self.include_private = include_private
        self.include_dunder = include_dunder
        self._style = style

        if include_all:
            self.include_docs = True
            self.include_mro = True
            self.include_method = True
            self.include_property = True
            self.include_attribute = True
            self.include_private = True

        # Setup line length
        if line_length is None:
            try:
                self._linelength = os.get_terminal_size().columns + line_length_offset
            except OSError:
                self._linelength = 88
        elif isinstance(line_length, (int, float)):
            self._linelength = max(int(line_length), 9)
        else:
            raise ValueError("Use different line_length")

        # Textwrap
        self._text_wrapper = TextWrapper(
            width=self._linelength - 4,
            initial_indent="",
            subsequent_indent="",
            tabsize=4,
            break_long_words=True,
            max_lines=max(max_textwrap_lines, 1),
        )

        # Output
        self._inspect_output = self._make_output()

    def __str__(self) -> str:
        return self._inspect_output.make_table()

    # 00. Support
    # -----------------------------------------------------------
    # @deprecated
    def _long_list_terminal_size(self, long_list: list) -> list:
        ll = ListExt(long_list).wrap_to_column(
            self._linelength, margin=4, transpose=True
        )
        return list(ll)

    # 01. Signature
    # -----------------------------------------------------------
    def _make_title(self) -> str:
        title_str = (
            str(self.obj)
            if (
                _inspect.isclass(self.obj)
                or callable(self.obj)
                or _inspect.ismodule(self.obj)
            )
            else str(type(self.obj))
        )
        return title_str

    def _get_signature_prefix(self) -> str:
        # signature prefix
        if _inspect.isclass(self.obj):
            return "class"
        elif _inspect.iscoroutinefunction(self.obj):
            return "async def"
        elif _inspect.isfunction(self.obj):
            return "def"
        return ""

    # @deprecated
    def get_parameters(self) -> list[str] | None:
        try:
            sig = _inspect.signature(self.obj)
        except (ValueError, AttributeError, TypeError):
            return None
        return [str(x) for x in sig.parameters.values()]

    def _get_func_signature(self, func: Callable[P, R]) -> list[str]:
        overloads = list(get_overloads(func))
        if len(overloads) < 1:
            return [
                f"{self._get_signature_prefix()} {func.__name__}{_inspect.signature(func)}"
            ]
        return [
            f"{self._get_signature_prefix()} {x.__name__}{_inspect.signature(x)}"
            for x in overloads
        ]

    # @deprecated
    def _make_signature(self) -> list[str]:
        try:
            # if isinstance(self.obj, Callable):
            if _inspect.isfunction(self.obj):
                funcs = [
                    self._text_wrapper.wrap(x)
                    for x in self._get_func_signature(self.obj)
                ]
                return ListExt(funcs).flatten()
            return self._text_wrapper.wrap(
                f"{self._get_signature_prefix()} {self.obj.__name__}{_inspect.signature(self.obj)}"
            )
        #    not class, func | not type   | is module
        except (ValueError, AttributeError, TypeError):
            return self._text_wrapper.wrap(repr(self.obj))

    @property
    def obj_signature(self) -> _TitleSignature:
        """Object's title and signature"""
        title: str = self._make_title()
        sig: list[str] = []
        try:
            if _inspect.isfunction(self.obj):
                sig.extend(self._get_func_signature(self.obj))
            sig.append(
                f"{self._get_signature_prefix()} {self.obj.__name__}{_inspect.signature(self.obj)}"
            )
        #    not class, func | not type   | is module
        except (ValueError, AttributeError, TypeError):
            sig.append(repr(self.obj))

        return _TitleSignature(title=title, signature=sig)

    # 02. Docstring
    # -----------------------------------------------------------
    @property
    def obj_docs(self) -> _Docstring:
        """Object's docstring"""
        docs: str | None = _inspect.getdoc(self.obj)

        if docs is None:
            return _Docstring("")
        return _Docstring(docs=docs)

    # @deprecated
    def _get_docs(self) -> str:
        docs: str | None = _inspect.getdoc(self.obj)

        if docs is None:
            return ""

        # Get docs and get first paragraph
        # doc_lines: list[str] = [x.strip() for x in docs.splitlines()]
        doc_lines = []
        for line in docs.splitlines():
            if len(line) < 1:
                break
            doc_lines.append(line.strip())

        return text_shorten(" ".join(doc_lines), width=self._linelength - 4, tabsize=4)

    # 03. MRO/Bases
    # -----------------------------------------------------------
    @property
    def obj_mro(self) -> _MRO:
        """Object's MRO, bases"""
        if isinstance(self.obj, type):
            return _MRO(mro=self.obj.__mro__[::-1])
        return _MRO(mro=type(self.obj).__mro__[::-1])

    @property
    def obj_bases(self) -> _MRO:
        """Object's MRO, bases"""
        return self.obj_mro

    # @deprecated
    def _get_mro(self) -> tuple[type, ...]:
        """Get MRO in reverse and subtract <class 'object'>"""
        if isinstance(self.obj, type):
            return self.obj.__mro__[::-1][1:]
        return type(self.obj).__mro__[::-1][1:]

    # @deprecated
    def _make_mro_data(self) -> list[str]:
        mro = [
            f"- {i:02}. {x.__module__}.{x.__name__}"
            for i, x in enumerate(self._get_mro(), start=1)
        ]
        mod_chunk = self._long_list_terminal_size(mro)

        # return [text_shorten(x, self._linelength - 4) for x in mod_chunk]
        return mod_chunk

    # 04. Class's members
    # -----------------------------------------------------------
    def _get_obj_member(self) -> ClassMembersResult:
        # if _inspect.isclass(self.obj) or inspect.ismodule(self.obj):
        if _inspect.isclass(self.obj):
            tmpcls = type(
                "tmpcls",
                (
                    self.obj,
                    GetClassMembersMixin,
                ),
                {},
            )
        else:
            tmpcls = type(
                "tmpcls",
                (
                    type(self.obj),
                    GetClassMembersMixin,
                ),
                {},
            )
        med_prop = tmpcls._get_members(  # type: ignore
            dunder=False, private=self.include_private
        )

        try:
            # If self.obj is a subclass of GetClassMembersMixin
            _mro = getattr(
                self.obj, "__mro__", getattr(type(self.obj), "__mro__", None)
            )
            if GetClassMembersMixin in _mro:  # type: ignore
                return med_prop  # type: ignore
        except AttributeError:  # Not a class
            pass
        med_prop.__delitem__(GetClassMembersMixin.__name__)
        return med_prop  # type: ignore

    @property
    def obj_member(self) -> _Member:
        """Object's members"""
        try:
            mem = self._get_obj_member()
            return _Member(mem.flatten_value())
        except (TypeError, AttributeError):
            return _Member(ClassMembers())

    # @deprecated
    def _get_method_property(self) -> ClassMembersResult:
        return self._get_obj_member()

    # Attribute
    @staticmethod
    def _is_real_attribute(obj: Any) -> bool:
        """
        Not method, classmethod, staticmethod, property
        """
        if callable(obj):
            return False
        if isinstance(obj, staticmethod):
            return False
        if isinstance(obj, classmethod):
            return False
        if isinstance(obj, property):
            return False
        return True

    def _get_attributes(self) -> list[tuple[str, Any]]:
        # Get attributes
        cls_dict = getattr(self.obj, "__dict__", None)
        cls_slots = getattr(self.obj, "__slots__", None)
        out = []

        # Check if __dict__ exist and len(__dict__) > 0
        if cls_dict is not None and len(cls_dict) > 0:
            if self.include_private:
                out = [
                    (k, v)
                    for k, v in self.obj.__dict__.items()
                    if self._is_real_attribute(v)
                ]
            else:
                out = [
                    (k, v)
                    for k, v in self.obj.__dict__.items()
                    if not k.startswith("_") and self._is_real_attribute(v)
                ]

        # Check if __slots__ exist and len(__slots__) > 0
        elif cls_slots is not None and len(cls_slots) > 0:
            if self.include_private:
                out = [
                    (x, getattr(self.obj, x))
                    for x in self.obj.__slots__  # type: ignore
                    if self._is_real_attribute(getattr(self.obj, x))
                ]
            else:
                out = [
                    (x, getattr(self.obj, x))
                    for x in self.obj.__slots__  # type: ignore
                    if not x.startswith("_")
                    and self._is_real_attribute(getattr(self.obj, x))
                ]

        return out

    def _handle_attributes_for_output(
        self, attr_list: list[tuple[str, Any]]
    ) -> list[str]:
        return [
            text_shorten(f"- {x[0]} = {x[1]}", self._linelength - 4) for x in attr_list
        ]

    # Output
    # -----------------------------------------------------------
    # @deprecated
    def _make_output(self) -> OneColumnTableMaker:
        table = OneColumnTableMaker(self._linelength, style=self._style)  # type: ignore
        body: list[str] = []

        # Signature
        title = self._make_title()
        table.add_title(title)
        sig = self._make_signature()
        if table._title == "":  # Title too long
            _title = [title]
            _title.extend(sig)
            table.add_paragraph(_title)
        else:
            table.add_paragraph(sig)

        # Docstring
        docs = self._get_docs()
        if len(docs) > 0 and self.include_docs:
            body.extend(["Docstring:", docs])

        # Class bases
        clsbases = self._make_mro_data()
        if len(clsbases) > 0 and self.include_mro:
            body.extend(["", f"Bases (Len: {len(self._get_mro())}):"])
            body.extend(clsbases)

        # Method & Property
        try:
            method_n_properties = (
                self._get_method_property().flatten_value().pack().sort()
            )
            if self.include_method:
                ml = [
                    text_shorten(f"- {x}", self._linelength - 4)
                    for x in method_n_properties.methods
                ]
                if len(ml) > 0:
                    head = ["", f"Methods (Len: {len(ml)}):"]
                    head.extend(self._long_list_terminal_size(ml))
                    body.extend(head)
            if self.include_property:
                pl = [
                    text_shorten(
                        f"- {x} = {getattr(self.obj, x, None)}", self._linelength - 4
                    )
                    for x in method_n_properties.properties
                ]
                if len(pl) > 0:
                    head = ["", f"Properties (Len: {len(pl)}):"]
                    head.extend(pl)
                    body.extend(head)
        except (TypeError, AttributeError):
            pass

        # Attribute
        attrs = self._get_attributes()
        if len(attrs) > 0 and self.include_attribute:
            body.extend(["", f"Attributes (Len: {len(attrs)}):"])
            body.extend(self._handle_attributes_for_output(attr_list=attrs))

        # Add to table
        table.add_paragraph(body)

        return table

    def _make_output_2(self) -> OneColumnTableMaker:
        # Prep
        # components: list[InspectComponents] = [
        #     self.obj_signature,
        #     self.obj_docs,
        #     self.obj_mro,
        #     self.obj_member,
        # ]
        table = OneColumnTableMaker(self._linelength, style=self._style)  # type: ignore
        body: list[str] = []

        # Signature
        title = self.obj_signature.title
        table.add_title(title)
        if table._title == "":  # Title too long
            _title = [title]
            _title.extend(self.obj_signature.signature)
            table.add_paragraph(_title)
        else:
            table.add_paragraph(self.obj_signature.signature)

        # Docstring
        if self.include_docs:
            body.extend(self.obj_docs.make_output())

        # Class bases
        if self.include_mro:
            body.extend(self.obj_mro.make_output(width=self._linelength))

        # Method & Property
        body.extend(
            self.obj_member.make_output(
                width=self._linelength,
                obj=self.obj,
                include_method=self.include_method,
                include_property=self.include_property,
            )
        )

        # Attribute
        attrs = self._get_attributes()
        if len(attrs) > 0 and self.include_attribute:
            body.extend(["", f"Attributes (Len: {len(attrs)}):"])
            body.extend(self._handle_attributes_for_output(attr_list=attrs))

        # Add to table
        table.add_paragraph(body)

        return table


# Partial
# ---------------------------------------------------------------------------
inspect_all = partial(Inspector, line_length=None, include_all=True)  # type: ignore



"""
Absufyu: Software
-----------------
Software, pyinstaller related stuff

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    # Function
    "get_system_info",
    "get_pyinstaller_exe_dir",
    "get_pyinstaller_resource_path",
    "PyinstallerHelper",
    "HWIDgen",
    "LicenseKeySystem",
    "BasicSoftwareProtection",
    # Support
    "SystemInfo",
    "LicenseKey",
    "PyinstallerHiddenImportPreset",
]


# Library
# ---------------------------------------------------------------------------
import base64
import hashlib
import hmac
import json
import os
import platform
import socket
import subprocess
import sys
import uuid
from datetime import datetime
from pathlib import Path
from typing import Literal, NamedTuple, TypedDict

from absfuyu.core.baseclass import BaseClass
from absfuyu.core.docstring import versionadded, versionchanged
from absfuyu.dxt import Text
from absfuyu.tools.converter import Base64EncodeDecode
from absfuyu.util import stop_after_day


# MARK: System Info
# ---------------------------------------------------------------------------
class SystemInfo(NamedTuple):
    """System info"""

    os_type: str | Literal["Windows", "Linux", "Darwin"]
    os_kernel: str
    os_arch: str
    system_name: str


def get_system_info() -> SystemInfo:
    """
    Returns the current operating system info.

    The function attempts to retrieve the computer name using `platform.node()`.
    If that fails or returns an empty string, it falls back to environment variables
    or `socket.gethostname()` as a last resort, depending on the OS.

    Returns
    -------
    SystemInfo
        A tuple containing:
        - OS name (e.g., 'Windows', 'Linux', 'Darwin')
        - OS kernel (version)
        - OS arch
        - Computer name (hostname)
    """
    os_name = platform.system()

    # Get computer name
    try:
        computer_name = platform.node()
        if not computer_name:
            raise ValueError("Empty name")
    except ValueError:
        if os_name == "Windows":
            computer_name = os.environ.get("COMPUTERNAME", "Unknown")
        else:
            computer_name = os.environ.get("HOSTNAME", socket.gethostname())

    # Return
    return SystemInfo(
        os_name, platform.version(), platform.machine().lower(), computer_name
    )


# MARK: Pyinstaller
# ---------------------------------------------------------------------------
@versionchanged("5.6.1", "Fixed behavior")
def get_pyinstaller_exe_dir() -> Path:
    """
    Returns the directory where the current script or executable resides.

    This function is useful for locating resources relative to the running script or
    bundled executable (e.g., when using PyInstaller). It checks if the script is
    running in a "frozen" state (as an executable), and returns the appropriate
    directory accordingly.

    Returns
    -------
    Path
        A `pathlib.Path` object representing the directory containing the
        executable or the current script file.
    """
    if getattr(sys, "frozen", False):
        return Path(sys.executable).parent
    else:
        return Path(sys.argv[0]).resolve().parent


def get_pyinstaller_resource_path(relative_path: str) -> Path:
    r"""
    Get the absolute path to a resource file, compatible with both development
    environments and PyInstaller-packaged executables.

    When running from a PyInstaller bundle, this function resolves the path relative
    to the temporary ``_MEIPASS`` folder. During normal execution, it resolves the path
    relative to the current script's directory.

    Parameters
    ----------
    relative_path : str
        Relative path to the resource file or directory.

    Returns
    -------
    Path
        A `pathlib.Path` object pointing to the absolute location of the resource.


    Example:
    --------
    >>> get_pyinstaller_resource_path("assets/logo.png")
    <path>\assets\logo.png
    """
    if hasattr(sys, "_MEIPASS"):
        # PyInstaller temp folder
        base_path = Path(getattr(sys, "_MEIPASS"))  # type: ignore[attr-defined]
    else:
        base_path = Path(__file__).resolve().parent

    return base_path / relative_path


class PyinstallerHiddenImportPreset:
    """
    pyinstaller hidden import preset (library preset)

    Example:
    --------
    >>> PyinstallerHelper(...).add_hidden_import(*PyinstallerHiddenImportPreset.ABSFUYU)
    """

    ABSFUYU = ["absfuyu"]
    DF = ["pandas", "numpy", "openpyxl", "xlsxwriter"]  # DataFrame
    VISUAL = ["rich", "tqdm"]


@versionadded("5.11.0")
class PyinstallerHelper(BaseClass):
    """pyinstaller helper"""

    def __init__(
        self,
        path_to_file: str | Path,
        relative_to_cwd: bool = True,
        console: bool = True,
        onefile: bool = False,
        noconfirm: bool = False,
    ) -> None:
        """
        pyinstaller cmd helper

        Parameters
        ----------
        path_to_file : str | Path
            Path to .py file to make .exe

        relative_to_cwd : bool, optional
            Is the file relative to cwd, by default True

        console : bool, optional
            Include console, by default True

        onefile : bool, optional
            Convert into one file, by default False

        noconfirm : bool, optional
            No confirmation, by default False
        """
        self.source_path = Path(path_to_file)
        self.relative_to_cwd = relative_to_cwd
        self.console = console
        self.onefile = onefile
        self.noconfirm = noconfirm

        if self.relative_to_cwd:
            # rel = self.source_path.relative_to(Path.cwd())
            # self._base_cmd = ["pyinstaller", f"'.\\{Path('.').joinpath(rel)}'"]
            self._base_cmd = ["pyinstaller", f"'.\\{self.source_path.relative_to(Path.cwd())}'"]
        else:
            self._base_cmd = ["pyinstaller", f"'{self.source_path.resolve()}'"]

        self._hidden_import = []
        self._icon = ""

    def add_hidden_import(self, *library: str) -> None:
        """
        Add hidden import (library)
        """
        if len(self._hidden_import) < 1:
            self._hidden_import = list(library)
        else:
            self._hidden_import.extend(list(library))

    def add_icon(self, path_to_icon: str | Path, *, relative_to_cwd: bool | None = None) -> None:
        """
        Add icon to .exe

        Parameters
        ----------
        path_to_icon : str | Path
            Path to icon file

        relative_to_cwd : bool | None, optional
            Use boolean value to overwrite relative_to_cwd option of the main engine, by default None
        """
        p = Path(path_to_icon)
        use_relative = self.relative_to_cwd if relative_to_cwd is None else relative_to_cwd

        if use_relative:
            rel = p.relative_to(Path.cwd())
            # ensure it always shows with ./ prefix
            # self._icon = f"'{Path('.').joinpath(rel)}'"
            self._icon = f"'.\\{rel}'"
        else:
            self._icon = f"'{p.resolve()}'"

    def export_cmd(self) -> str:
        """
        Export pyinstaller cmd

        Returns
        -------
        str
            pyinstaller command
        """
        cmd = self._base_cmd

        # Hidden import
        if len(self._hidden_import) > 0:
            dat = (f"--hidden-import={x}" for x in list(set(self._hidden_import)))
            cmd.append(" ".join(dat))

        # Console
        if not self.console:
            cmd.append("--noconsole")

        # One file
        if self.onefile:
            cmd.append("--onefile")

        # No confirm
        if self.noconfirm:
            cmd.append("--noconfirm")

        # Icon
        if len(self._icon) > 0:
            cmd.append(self._icon)

        return " ".join(cmd)


# MARK: Key System
# ---------------------------------------------------------------------------
class HWIDgen(BaseClass):
    """
    Generate Hardware ID (HWID)


    Example:
    --------
    >>> HWIDgen.generate()
    """

    def __init__(self) -> None:
        pass

    @classmethod
    def generate(cls) -> str:
        """Generate HWID for current system"""
        os_type = platform.system().lower()
        if os_type == "windows":
            return cls._get_windows_hwid()
        elif os_type == "linux":
            return cls._get_linux_hwid()
        else:
            return cls._get_hwid_mac()

    @staticmethod
    def _get_hwid_mac() -> str:
        """HWID: MAC address"""
        mac = uuid.getnode()  # 48-bit MAC address
        mac_str = ":".join(("%012X" % mac)[i : i + 2] for i in range(0, 12, 2))
        hwid = hashlib.sha256(mac_str.encode()).hexdigest()
        return hwid

    @staticmethod
    def _get_windows_hwid() -> str:
        try:
            # Get BIOS serial number
            bios = (
                subprocess.check_output("wmic bios get serialnumber", shell=True)
                .decode()
                .split("\n")[1]
                .strip()
            )

            # Get Motherboard serial
            board = (
                subprocess.check_output("wmic baseboard get serialnumber", shell=True)
                .decode()
                .split("\n")[1]
                .strip()
            )

            # Get Disk serial
            disk = (
                subprocess.check_output("wmic diskdrive get serialnumber", shell=True)
                .decode()
                .split("\n")[1]
                .strip()
            )

            raw = bios + board + disk
            hwid = hashlib.sha256(raw.encode()).hexdigest()
            return hwid
        except Exception as e:
            return f"Error getting HWID: {e}"

    @staticmethod
    def _get_linux_hwid() -> str:
        try:
            disk_info = subprocess.check_output(
                "udevadm info --query=all --name=/dev/sda", shell=True
            ).decode()
            serial = ""
            for line in disk_info.splitlines():
                if "ID_SERIAL_SHORT" in line:
                    serial = line.split("=")[1]
                    break

            mac = uuid.getnode()
            mac_str = ":".join(("%012X" % mac)[i : i + 2] for i in range(0, 12, 2))

            raw = serial + mac_str
            hwid = hashlib.sha256(raw.encode()).hexdigest()
            return hwid
        except Exception as e:
            return f"Error getting HWID: {e}"


class LicenseKey(TypedDict):
    name: str
    expiry: str
    signature: str


class LicenseKeySystem(BaseClass):
    def __init__(self, name: str, expiry: str, secret_key: str) -> None:
        """
        License Key implementation

        Parameters
        ----------
        name : str
            Name of license holder

        expiry : str
            Expiry date (in yyyy-mm-dd format)

        secret_key : str
            Secret key to make license key
        """
        self._name = name
        self._expiry = expiry
        self._secret = secret_key.encode("utf-8")

    # Make key
    def make_key(self, *, hwid_overwrite: str | None = None) -> str:
        """
        Make a license key in these following steps:
        1. Generate HWID
        2. Combine name, expiry, HWID -> sign it
        3. Base64-encode the signature
        4. Store name, expiry, signature in JSON -> Base64 -> Hex
        5. Divide by 30 chars per line
        6. Wrap in BEGIN/END KEY

        Parameters
        ----------
        hwid_overwrite : str | None, optional
            Overwrite the HWID, by default None

        Returns
        -------
        str
            License key
        """
        # Prepare
        hwid = (
            HWIDgen.generate() if hwid_overwrite is None else hwid_overwrite
        )  # Get HWID
        msg_data = f"{self._name}|{self._expiry}|{hwid}"  # Make msg
        signature = hmac.new(self._secret, msg_data.encode(), hashlib.sha256).digest()
        encoded_sig = base64.urlsafe_b64encode(signature).decode()

        # Key format
        key: LicenseKey = {
            "name": self._name,
            "expiry": self._expiry,
            "signature": encoded_sig,
        }
        # This convert to .json -> Base64 -> Hex
        encoded_key = Text(Base64EncodeDecode.encode(json.dumps(key))).to_hex(raw=True)
        output_key = (
            "BEGIN KEY".center(30, "=")
            + "\n"
            + "\n".join(Text(encoded_key).divide(30))
            + "\n"
            + "END KEY".center(30, "=")
        )
        return output_key

    # Check key
    @staticmethod
    def _parse_license_key(license_key: str) -> LicenseKey:
        """
        Parse a formatted license key

        Parameters
        ----------
        license_key : str
            License key

        Returns
        -------
        LicenseKey
            Parsed license key
        """
        try:
            lines = license_key.strip().split("\n")[1:-1]  # Remove BEGIN/END KEY lines
            raw_hex = "".join(lines)
            decoded_json = Base64EncodeDecode.decode(
                bytes.fromhex(raw_hex).decode("utf-8")
            )
            return json.loads(decoded_json)
        except Exception as e:
            raise ValueError("Invalid license key format") from e

    @classmethod
    def verify_license(
        cls,
        license_key: str,
        secret_key: str | None = None,
        hwid_overwrite: str | None = None,
    ) -> bool:
        """
        Verify a license key

        Parameters
        ----------
        license_key : str
            License key

        secret_key : str | None, optional
            Secret key, by default None

        hwid_overwrite : str | None, optional
            HWID, by default None

        Returns
        -------
        bool
            _description_
        """

        # Prep
        secret = "" if secret_key is None else secret_key
        hwid = HWIDgen.generate() if hwid_overwrite is None else hwid_overwrite
        parsed_license_key = cls._parse_license_key(license_key)

        try:
            msg_data = (
                f"{parsed_license_key['name']}|{parsed_license_key['expiry']}|{hwid}"
            )

            expected_sig = hmac.new(
                secret.encode("utf-8"), msg_data.encode(), hashlib.sha256
            ).digest()
            expected_encoded_sig = base64.urlsafe_b64encode(expected_sig).decode()

            return parsed_license_key["signature"] == expected_encoded_sig
        except Exception:
            return False


# MARK: Software
# ---------------------------------------------------------------------------
class BasicSoftwareProtection(BaseClass):
    """
    Basic software protection

    This check valid license before run any app. Recommended to put at start of the code

    Usage:
    ------
    >>> t = BasicSoftwareProtection(get_pyinstaller_exe_dir())
    >>> t.add_secret("Test Key")
    >>> t.check_valid_license()
    """

    def __init__(
        self,
        cwd: str | Path,
        name: str | None = None,
        version: str | None = None,
        author: str | None = None,
        author_email: str | None = None,
    ) -> None:
        """
        Basic software protection.

        Parameters
        ----------
        cwd : str | Path
            Current working directory

        name : str | None, optional
            Name of the software, by default None

        version : str | None, optional
            Version of the software, by default None

        author : str | None, optional
            Author of the software, by default None

        author_email : str | None, optional
            Author's email of the software, by default None
        """
        self._cwd = Path(cwd)
        self._author = "" if author is None else author
        self._author_email = "" if author_email is None else author_email
        self._software_name = "" if name is None else name
        self._software_version = "" if version is None else version
        self._secret = ""

    # Metadata
    @property
    def cwd(self) -> Path:
        """Current working directory"""
        return self._cwd

    @property
    def software_name(self) -> str:
        """Name of the software"""
        return self._software_name

    @software_name.setter
    def software_name(self, value: str) -> None:
        # Logic to validate name
        self._software_name = value

    @property
    def author(self) -> str:
        """Author of the software"""
        return self._author

    @property
    def version(self) -> str:
        """Version of the software"""
        return self._software_version

    # Protection
    def add_secret(self, secret: str) -> None:
        """
        Add secret

        Parameters
        ----------
        secret : str
            secret
        """
        self._secret = secret

    def check_valid_license(self, generate_helper: bool = True) -> None:
        """
        Check for valid license
        (``.zlic`` file in the same directory as the script that runs this code)

        Parameters
        ----------
        generate_helper : bool, optional
            Generate a helper file (``license.helper``),
            which can be used to make license key,
            by default ``True``

        Raises
        ------
        SystemExit
            License file not found!

        SystemExit
            Invalid license key format!

        SystemExit
            Invalid license!
        """
        try:
            # Get license file
            license_file = list(self._cwd.glob("*.zlic"))[0]
            with license_file.open() as f:
                # Load data
                data = "".join(f.readlines())
        except IndexError:
            if generate_helper:
                self.generate_license_helper()
            raise SystemExit("License file not found!")
        except ValueError:
            raise SystemExit("Invalid license key format!")
        else:
            # Verify license
            if LicenseKeySystem.verify_license(data, secret_key=self._secret):
                parsed_date = datetime.strptime(
                    LicenseKeySystem._parse_license_key(data)["expiry"], "%Y-%m-%d"
                )
                stop_after_day(
                    parsed_date.year,
                    parsed_date.month,
                    parsed_date.day,
                    custom_msg="License expired!",
                )
            else:  # Invalid license
                raise SystemExit("Invalid license!")

    @versionadded("5.9.0")
    def check_valid_license_gui(
        self, title: str = "WARNING", message: str = "INVALID LICENSE!", generate_helper: bool = True
    ) -> None:
        """
        Check for valid license
        (``.zlic`` file in the same directory as the script that runs this code)
        but will pop up a GUI when invalid license.

        Parameters
        ----------
        title : str, optional
            Title of the GUI, by default "WARNING"

        message : str, optional
            Message in the GUI, by default "INVALID LICENSE!"

        generate_helper : bool, optional
            Generate a helper file (``license.helper``),
            which can be used to make license key,
            by default ``True``

        Raises
        ------
        SystemExit
            Invalid license
        """
        try:
            self.check_valid_license(generate_helper=generate_helper)
        except SystemExit:
            from tkinter import messagebox

            messagebox.showwarning(title, message, icon="error")
            raise SystemExit("Invalid license")

    # Make key
    def _make_key(self, name: str, expiry: str, secret: str) -> None:
        """
        Generate license key in the same directory as the script that runs this code.

        Parameters
        ----------
        name : str
            Name of license's holder

        expiry : str
            Expiry date in format "yyyy-mm-dd"

        secret : str
            Secret string
        """
        path = self._cwd.joinpath("license.zlic")
        engine = LicenseKeySystem(name, expiry, secret)
        with path.open("w", encoding="utf-8") as f:
            f.write(engine.make_key())

    def generate_license_helper(self) -> None:
        """Gather HWID and make it into a file"""
        path = self._cwd.joinpath("license.helper")
        with path.open("w", encoding="utf-8") as f:
            f.write(HWIDgen.generate())



"""
ABSFUYU CLI
-----------
Do

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["do_group"]


# Library
# ---------------------------------------------------------------------------
import subprocess

import click

from absfuyu import __title__
from absfuyu.cli.color import COLOR
from absfuyu.core import __package_feature__
from absfuyu.tools.shutdownizer import ShutDownizer
from absfuyu.util.path import Directory
from absfuyu.version import PkgVersion


# CLI
# ---------------------------------------------------------------------------
@click.command()
@click.option(
    "--force_update/--no-force-update",
    "-F/-f",
    "force_update",
    type=bool,
    default=True,
    show_default=True,
    help="Update the package",
)
def update(force_update: bool) -> None:
    """Update the package to latest version"""
    click.echo(f"{COLOR['green']}")
    AbsfuyuPackage = PkgVersion(
        package_name=__title__,
    )
    AbsfuyuPackage.check_for_update(force_update=force_update)


@click.command()
@click.argument("pkg", type=click.Choice(__package_feature__))
def install(pkg: str) -> None:
    """Install absfuyu's extension"""
    cmd = f"pip install -U absfuyu[{pkg}]".split()
    try:
        subprocess.run(cmd)
    except Exception:
        try:
            cmd2 = f"python -m pip install -U absfuyu[{pkg}]".split()
            subprocess.run(cmd2)
        except Exception:
            click.echo(f"{COLOR['red']}Unable to install absfuyu[{pkg}]")
        else:
            click.echo(f"{COLOR['green']}absfuyu[{pkg}] installed")
    else:
        click.echo(f"{COLOR['green']}absfuyu[{pkg}] installed")


@click.command(name="unzip")
@click.argument("dir", type=str)
def unzip_files_in_dir(dir: str) -> None:
    """Unzip every files in directory"""

    engine = Directory(dir)
    engine.decompress()
    click.echo(f"{COLOR['green']}Done!")


@click.command(name="shutdown")
def os_shutdown() -> None:
    """Shutdown"""
    engine = ShutDownizer()
    engine.shutdown()


@click.command(name="organize")
@click.argument("dir", type=str)
def organize_directory(dir: str) -> None:
    """Organize a directory"""
    engine = Directory(dir)
    engine.organize()
    click.echo(f"{COLOR['green']}Done!")


@click.group(name="do")
def do_group() -> None:
    """Perform functionalities"""
    pass


do_group.add_command(update)
do_group.add_command(install)
do_group.add_command(unzip_files_in_dir)
do_group.add_command(os_shutdown)
do_group.add_command(organize_directory)



"""
Absfuyu: Data Extension
-----------------------
list extension

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["ListExt"]


# Library
# ---------------------------------------------------------------------------
import operator
import random
from collections import Counter
from collections.abc import Callable, Iterable
from heapq import heapreplace
from itertools import accumulate, chain, count, groupby, zip_longest
from typing import Any, Literal, Self, TypeVar, cast, overload

from absfuyu.core.baseclass import GetClassMembersMixin
from absfuyu.core.docstring import deprecated, versionadded, versionchanged
from absfuyu.util import set_min_max

# Type
# ---------------------------------------------------------------------------
T = TypeVar("T")
R = TypeVar("R")  # Return type - Can be anything

# Class
# ---------------------------------------------------------------------------
class ListExt(GetClassMembersMixin, list):
    """
    ``list`` extension

    >>> # For a list of new methods
    >>> ListExt.show_all_methods()
    """

    # Deprecated
    @deprecated("5.2.0", reason="Use ListExt.apply(str) instead")
    def stringify(self) -> Self:
        """
        Convert all item in ``list`` into string

        Returns
        -------
        Self
            A list with all items with type <str`>


        Example:
        --------
        >>> test = ListExt([1, 1, 1, 2, 2, 3])
        >>> test.stringify()
        ['1', '1', '1', '2', '2', '3']
        """
        return self.__class__(map(str, self))

    # Info
    def head(self, number_of_items: int = 5) -> list:
        """
        Show first ``number_of_items`` items in ``ListExt``

        Parameters
        ----------
        number_of_items : int
            | Number of items to shows at once
            | (Default: ``5``)

        Returns
        -------
        list
            Filtered list


        Example:
        --------
        >>> ListExt(range(10)).head(2)
        [0, 1]
        """
        number_of_items = int(set_min_max(number_of_items, min_value=0, max_value=len(self)))
        return self[:number_of_items]

    def tail(self, number_of_items: int = 5) -> list:
        """
        Show last ``number_of_items`` items in ``ListExt``

        Parameters
        ----------
        number_of_items : int
            | Number of items to shows at once
            | (Default: ``5``)

        Returns
        -------
        list
            Filtered list


        Example:
        --------
        >>> ListExt(range(10)).tail(2)
        [8, 9]
        """
        number_of_items = int(set_min_max(number_of_items, min_value=0, max_value=len(self)))
        return self[::-1][:number_of_items][::-1]

    # Misc
    def apply(self, func: Callable[[Any], Any]) -> Self:
        """
        Apply function to each entry

        Parameters
        ----------
        func : Callable
            Callable function

        Returns
        -------
        Self
            ListExt


        Example:
        --------
        >>> test = ListExt([1, 2, 3])
        >>> test.apply(str)
        ['1', '2', '3']
        """
        # return self.__class__(map(func, self))
        return self.__class__(func(x) for x in self)

    def sorts(self, reverse: bool = False) -> Self:
        """
        Sort all items (with different type) in ``list``

        Parameters
        ----------
        reverse : bool
            - ``True`` then sort in descending order
            - ``False`` then sort in ascending order (default value)

        Returns
        -------
        Self
            A sorted list


        Example:
        --------
        >>> test = ListExt([9, "abc", 3.5, "aaa", 1, 1.4])
        >>> test.sorts()
        [1, 9, 'aaa', 'abc', 1.4, 3.5]
        """
        lst = self.copy()
        type_weights: dict = {}
        for x in lst:
            if type(x) not in type_weights:
                type_weights[type(x)] = len(type_weights)
        # logger.debug(f"Type weight: {type_weights}")

        output = sorted(lst, key=lambda x: (type_weights[type(x)], str(x)), reverse=reverse)

        # logger.debug(output)
        return self.__class__(output)

    @overload
    def freq(self) -> dict: ...  # type: ignore

    @overload
    def freq(
        self,
        sort: bool = False,
        num_of_first_char: int | None = None,
    ) -> dict: ...

    @overload
    def freq(
        self,
        sort: bool = False,
        num_of_first_char: int | None = None,
        appear_increment: Literal[True] = ...,
    ) -> list[int]: ...

    def freq(
        self,
        sort: bool = False,
        num_of_first_char: int | None = None,
        appear_increment: bool = False,
    ) -> dict | list[int]:
        """
        Find frequency of each item in list

        Parameters
        ----------
        sort : bool, optional
            - ``True``: Sorts the output in ascending order
            - ``False``: No sort (default value)

        num_of_first_char : int | None, optional
            | Number of first character taken into account to sort, by default ``None``
            | Eg: ``num_of_first_char = 1``: first character of each item

        appear_increment : bool, optional
            Returns incremental index list of each item when sort,
            by default ``False``

        Returns
        -------
        dict
            A dict that show frequency

        list[int]
            Incremental index list


        Example:
        --------
        >>> test = ListExt([1, 1, 2, 3, 5, 5])
        >>> test.freq()
        {1: 2, 2: 1, 3: 1, 5: 2}

        >>> test = ListExt([1, 1, 2, 3, 3, 4, 5, 6])
        >>> test.freq(appear_increment=True)
        [2, 3, 5, 6, 7, 8]
        """
        if sort:
            data = self.sorts().copy()
        else:
            data = self.copy()

        if num_of_first_char is None:
            temp = Counter(data)
        else:
            max_char: int = min([len(str(x)) for x in data])
            # logger.debug(f"Max character: {max_char}")
            if num_of_first_char not in range(1, max_char):
                # logger.debug(f"Not in {range(1, max_char)}. Using default value...")
                temp = Counter(data)
            else:
                # logger.debug(f"Freq of first {num_of_first_char} char")
                temp = Counter([str(x)[:num_of_first_char] for x in data])

        try:
            times_appear = dict(sorted(temp.items()))
        except TypeError:
            times_appear = dict(self.__class__(temp.items()).sorts())
        # logger.debug(times_appear)

        if appear_increment:
            times_appear_increment: list[int] = list(accumulate(times_appear.values(), operator.add))
            # logger.debug(times_appear_increment)
            return times_appear_increment  # incremental index list
        else:
            return times_appear  # character frequency

    @versionchanged("5.2.0", reason="New ``recursive`` parameter")
    def flatten(self, recursive: bool = False) -> Self:
        """
        Flatten the list

        Parameters
        ----------
        recursive : bool
            Recursively flatten the list, by default ``False``

        Returns
        -------
        Self
            Flattened list


        Example:
        --------
        >>> test = ListExt([["test"], ["test", "test"], ["test"]])
        >>> test.flatten()
        ['test', 'test', 'test', 'test']
        """

        def instance_checking(item):
            if isinstance(item, str):
                return [item]
            if isinstance(item, Iterable):
                return item
            return [item]

        # Flatten
        list_of_list = (instance_checking(x) for x in self)
        flattened = self.__class__(chain(*list_of_list))

        # Recursive
        if recursive:

            def _condition(item) -> bool:
                if isinstance(item, str):
                    return False
                return isinstance(item, Iterable)

            while any(flattened.apply(_condition)):
                flattened = flattened.flatten()

        # Return
        return flattened

    @versionadded("5.2.0")
    def join(self, sep: str = " ", /) -> str:
        """
        Join every element in list to str (``str.join()`` wrapper).

        Parameters
        ----------
        sep : str, optional
            Separator between each element, by default ``" "``

        Returns
        -------
        str
            Joined list


        Example:
        --------
        >>> ListExt(["a", "b", "c"]).join()
        a b c

        >>> # Also work with non-str type
        >>> ListExt([1, 2, 3]).join()
        1 2 3
        """
        try:
            return sep.join(self)
        except TypeError:
            return sep.join(self.apply(str))

    @overload
    def numbering(self) -> Self: ...

    @overload
    def numbering(self, start: int | float = 0) -> Self: ...

    @overload
    def numbering(self, *, step: int | float = 1) -> Self: ...

    @overload
    def numbering(self, start: int | float = 0, step: int | float = 1) -> Self: ...

    @versionchanged("5.5.0", "Use itertools.count to wrap")
    def numbering(self, start: int | float = 0, step: int | float = 1) -> Self:
        """
        Number the item in list
        (``itertools.count`` wrapper)

        Parameters
        ----------
        start : int | float, optional
            Start from which number, by default ``0``

        step : int | float, optional
            Step, by default ``1``

        Returns
        -------
        Self
            Counted list


        Example:
        --------
        >>> test = ListExt([9, 9, 9])
        >>> test.numbering()
        [(0, 9), (1, 9), (2, 9)]
        """
        nums = count(start=start, step=step)
        # return self.__class__(enumerate(self, start=start))
        return self.__class__(zip(nums, self))

    @versionadded("5.3.0")  # no test case yet
    def transpose(self, fillvalue: Any | None = None, /) -> Self:
        """
        Transpose a list of iterable.

        Parameters
        ----------
        fillvalue : Any, optional
            A fill value, by default ``None``

        Returns
        -------
        Self | list[list[T]]
            Transposed list.


        Example:
        --------
        >>> ListExt([1, 1, 1, 1]).transpose()
        [(1, 1, 1, 1)]

        >>> ListExt([[1, 1, 1, 1], [1, 1, 1, 1]]).transpose()
        [(1, 1), (1, 1), (1, 1), (1, 1)]

        >>> ListExt([[1, 1, 1, 1], [1, 1, 1, 1], [1]]).transpose()
        [(1, 1, 1), (1, 1, None), (1, 1, None), (1, 1, None)]
        """
        try:
            return self.__class__(zip_longest(*self, fillvalue=fillvalue)).apply(list)
        except TypeError:  # Dimension of 1
            mod_dat = self.apply(lambda x: [x])
            # return self.__class__(zip_longest(*mod_dat, fillvalue=fillvalue)).apply(list)
            return mod_dat

    # Random
    def pick_one(self) -> Any:
        """
        Pick one random items from ``list``

        Returns
        -------
        Any
            Random value


        Example:
        --------
        >>> test = ListExt(["foo", "bar"])
        >>> test.pick_one()
        'bar'
        """
        if len(self) != 0:
            out = random.choice(self)
            # logger.debug(out)
            return out
        else:
            # logger.debug("List empty!")
            raise IndexError("List empty!")

    def get_random(self, number_of_items: int = 5) -> list:
        """
        Get ``number_of_items`` random items in ``ListExt``

        Parameters
        ----------
        number_of_items : int
            Number random of items, by default ``5``

        Returns
        -------
        list
            Filtered list
        """
        return [self.pick_one() for _ in range(number_of_items)]

    @versionadded("5.10.0")
    def shuffle(self) -> Self:
        random.shuffle(self)
        return self

    # Len
    @versionchanged("5.2.0", reason="Handle more type")
    def len_items(self) -> Self:
        """
        ``len()`` for every item in ``self``

        Returns
        -------
        Self
            List of ``len()``'ed value


        Example:
        --------
        >>> test = ListExt(["foo", "bar", "pizza"])
        >>> test.len_items()
        [3, 3, 5]
        """

        def _len(item: Any) -> int:
            try:
                return len(item)
            except TypeError:
                return len(str(item))

        return self.__class__([_len(x) for x in self])

    @versionchanged("5.2.0", reason="New ``recursive`` parameter")
    def mean_len(self, recursive: bool = False) -> float:
        """
        Average length of every item. Returns zero if failed (empty list, ZeroDivisionError).

        Parameters
        ----------
        recursive : bool
            Recursively find the average length of items in nested lists, by default ``False``

        Returns
        -------
        float
            Average length


        Example:
        --------
        >>> test = ListExt(["foo", "bar", "pizza"])
        >>> test.mean_len()
        3.6666666666666665
        """

        if recursive:
            dat = self.flatten(recursive=recursive)
        else:
            dat = self

        try:
            return sum(dat.len_items()) / len(dat)
        except ZeroDivisionError:
            return 0.0

    @versionadded("5.2.0")
    def max_item_len(self, recursive: bool = False) -> int:
        """
        Find the maximum length of items in the list.

        Parameters
        ----------
        recursive : bool
            Recursively find the maximum length of items in nested lists, by default ``False``

        Returns
        -------
        int
            Maximum length of items


        Example:
        --------
        >>> test = ListExt(["test", "longer_test"])
        >>> test.max_item_len()
        11

        >>> test = ListExt([["short"], ["longer_test"]])
        >>> test.max_item_len(recursive=True)
        11
        """
        if recursive:
            return cast(int, max(self.flatten(recursive=True).len_items()))
        return cast(int, max(self.len_items()))

    # Group/Unique
    def unique(self) -> Self:
        """
        Remove duplicates

        Returns
        -------
        Self
            Duplicates removed list


        Example:
        --------
        >>> test = ListExt([1, 1, 1, 2, 2, 3])
        >>> test.unique()
        [1, 2, 3]
        """
        return self.__class__(set(self))

    def group_by_unique(self) -> Self:
        """
        Group duplicated elements into list

        Returns
        -------
        Self
            Grouped value


        Example:
        --------
        >>> test = ListExt([1, 2, 3, 1, 3, 3, 2])
        >>> test.group_by_unique()
        [[1, 1], [2, 2], [3, 3, 3]]
        """
        # Old
        # out = self.sorts().slice_points(self.freq(appear_increment=True))
        # return __class__(out[:-1])

        # New
        temp = groupby(self.sorts())
        return self.__class__([list(g) for _, g in temp])

    def group_by_pair_value(self, max_loop: int = 3) -> list[list]:
        """
        Assume each ``list`` in ``list`` is a pair value,
        returns a ``list`` contain all paired value

        Parameters
        ----------
        max_loop : int
            Times to run functions (minimum: ``3``)

        Returns
        -------
        list[list]
            Grouped value


        Example:
        --------
        >>> test = ListExt([[1, 2], [2, 3], [4, 3], [5, 6]])
        >>> test.group_by_pair_value()
        [[1, 2, 3, 4], [5, 6]]

        >>> test = ListExt([[8, 3], [4, 6], [6, 3], [5, 2], [7, 2]])
        >>> test.group_by_pair_value()
        [[8, 3, 4, 6], [2, 5, 7]]

        >>> test = ListExt([["a", 4], ["b", 4], [5, "c"]])
        >>> test.group_by_pair_value()
        [['a', 4, 'b'], ['c', 5]]
        """

        iter = self.copy()

        # Init loop
        for _ in range(max(max_loop, 3)):
            temp: dict[Any, list] = {}
            # Make dict{key: all `item` that contains `key`}
            for item in iter:
                for x in item:
                    if temp.get(x, None) is None:
                        temp[x] = [item]
                    else:
                        temp[x].append(item)

            # Flatten dict.values
            temp = {k: list(set(chain(*v))) for k, v in temp.items()}

            iter = list(temp.values())

        return list(x for x, _ in groupby(iter))

    @versionadded("5.5.0")
    def split_equal(self, n: int, sort: bool = True) -> Self:
        """
        Try to equally split a list of number into ``n`` equal sum parts.

        **Note:** Element in list must be a number.

        Parameters
        ----------
        n : int
            Split into how many parts. Must be >= 1.

        sort : bool, optional
            Sort the instance before split, by default ``True``

        Returns
        -------
        Self
            Splitted (list[list])


        Example:
        --------
        >>> ListExt(range(1, 11)).split_equal(2)
        [[10, 7, 5, 4, 1], [9, 8, 6, 3, 2]]
        """

        # https://stackoverflow.com/a/61649667
        bins: list[list[int]] = [[0] for _ in range(max(n, 1))]
        if sort:
            # self = self.sorts(reverse=True)
            self = self.__class__(sorted(self, reverse=True))
        for x in self:
            least = bins[0]
            least[0] += x
            least.append(x)
            heapreplace(bins, least)
        return self.__class__(x[1:] for x in bins)

    # Slicing
    def slice_points(self, points: list[int]) -> Self:
        """
        Splits a list into sublists based on specified split points (indices).

        This method divides the original list into multiple sublists. The ``points``
        argument provides the indices at which the list should be split.  The resulting
        list of lists contains the sublists created by these splits. The original
        list is not modified.

        Parameters
        ----------
        points : list
            A list of integer indices representing the points at which to split
            the list. These indices are *exclusive* of the starting sublist
            but *inclusive* of the ending sublist.

        Returns
        -------
        Self | list[list]
            A list of lists, where each inner list is a slice of the original list
            defined by the provided split points.


        Example:
        --------
        >>> test = ListExt([1, 1, 2, 3, 3, 4, 5, 6])
        >>> test.slice_points([2, 5])
        [[1, 1], [2, 3, 3], [4, 5, 6]]
        >>> test.slice_points([0, 1, 2, 3, 4, 5, 6, 7, 8])
        [[], [1], [1], [2], [3], [3], [4], [5], [6]]
        >>> test.slice_points([])
        [[1, 1, 2, 3, 3, 4, 5, 6]]
        """
        points.append(len(self))
        data = self.copy()
        # return [data[points[i]:points[i+1]] for i in range(len(points)-1)]
        return self.__class__(data[i1:i2] for i1, i2 in zip([0] + points[:-1], points))

    @versionadded("5.1.0")
    def split_chunk(self, chunk_size: int, /) -> Self:
        """
        Split list into smaller chunks

        Parameters
        ----------
        chunk_size : int
            Chunk size, minimum: ``1``

        Returns
        -------
        Self | list[list]
            List of chunk


        Example:
        --------
        >>> ListExt([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]).split_chunk(5)
        [[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1]]
        """
        slice_points = list(range(0, len(self), max(chunk_size, 1)))[1:]
        return self.slice_points(slice_points)

    @versionadded("5.3.0")  # no test case yet
    def to_column(self, ncols: int, fillvalue: Any | None = None) -> Self:
        """
        Smart convert 1 dimension list to 2 dimension list,
        in which, number of columns = ``ncols``.

        Parameters
        ----------
        ncols : int
            Number of columns

        fillvalue : Any | None, optional
            Fill value, by default ``None``

        Returns
        -------
        Self
            Coulumned list.


        Example:
        --------
        >>> ins = ListExt(range(1, 20))

        >>> # Normal split chunk
        >>> ins.split_chunk(10)
        [
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            [11, 12, 13, 14, 15, 16, 17, 18, 19]
        ]

        >>> # Column split chunk
        >>> ins.to_column(10)
        [
            [1, 3, 5, 7, 9, 11, 13, 15, 17, 19],
            [2, 4, 6, 8, 10, 12, 14, 16, 18, None]
        ]
        """
        num_of_col = max(ncols, 1)
        len_cols = len(self.split_chunk(num_of_col))
        return self.split_chunk(len_cols).transpose(fillvalue)

    @overload
    def wrap_to_column(self, width: int, /) -> Self: ...

    @overload
    def wrap_to_column(
        self,
        width: int,
        /,
        *,
        margin: int = 4,
        sep: str = "",
        fill: str = " ",
        transpose: bool = False,
    ) -> Self: ...

    @versionchanged("5.3.0", reason="New `sep`, `fill`, `transpose` parameters")
    @versionadded("5.2.0")  # no test case yet
    def wrap_to_column(
        self,
        width: int,
        /,
        *,
        margin: int = 4,
        sep: str = "",
        fill: str = " ",
        transpose: bool = False,
    ) -> Self:
        """
        Arrange list[str] items into aligned text columns (for printing)
        with automatic column count calculation.

        Parameters
        ----------
        width : int
            Total available display width (must be >= ``margin``)

        margin : int, optional
            Reserved space for borders/padding, should be an even number, by default ``4``

        sep : str, optional
            Separator between each element, by default ``""``

        fill : str, optional
            Fill character for spacing, must have the length of 1, by default ``" "``

        transpose : bool, optional
            Smart transpose the columns, by default ``False``

        Returns
        -------
        Self
            New instance type[list[str]] with formatted column strings.


        Example:
        --------
        >>> items = ListExt(["apple", "banana", "cherry", "date"])
        >>> print(items.wrap_to_column(30))
        ['apple  banana cherry ', 'date   ']

        >>> items.wrap_to_column(15)
        ['apple  ', 'banana ', 'cherry ', 'date   ']
        """

        max_item_length = self.max_item_len() + max(len(sep), 1)
        available_width = max(width, 4) - max(margin, 0)  # Set boundary
        if len(fill) != 1:
            fill = " "

        # Calculate how many columns of text
        column_count = max(1, available_width // max_item_length) if max_item_length > 0 else 1

        # splitted_chunk: list[list[str]] = self.split_chunk(cols)
        # mod_chunk = self.__class__(
        #     [[x.ljust(max_name_len, " ") for x in chunk] for chunk in splitted_chunk]
        # ).apply(lambda x: "".join(x))

        def mod_item(item: list[str]) -> str:
            # Set width for str item and join them together
            return sep.join(x.ljust(max_item_length, fill) for x in item)

        if transpose:
            mod_chunk = self.to_column(column_count, fillvalue="").apply(mod_item)
        else:
            mod_chunk = self.split_chunk(column_count).apply(mod_item)

        return mod_chunk


class ListExt2(GetClassMembersMixin, list[T]):
    """
    ``list`` extension (with generic - W.I.P)

    >>> # For a list of new methods
    >>> ListExt2.show_all_methods()
    """

    # Info
    def head(self, number_of_items: int = 5, /) -> list[T]:
        """
        Show first ``number_of_items`` items in ``ListExt``

        Parameters
        ----------
        number_of_items : int
            | Number of items to shows at once
            | (Default: ``5``)

        Returns
        -------
        list
            Filtered list


        Example:
        --------
        >>> ListExt(range(10)).head(2)
        [0, 1]
        """
        number_of_items = int(set_min_max(number_of_items, min_value=0, max_value=len(self)))
        return self[:number_of_items]

    def tail(self, number_of_items: int = 5, /) -> list[T]:
        """
        Show last ``number_of_items`` items in ``ListExt``

        Parameters
        ----------
        number_of_items : int
            | Number of items to shows at once
            | (Default: ``5``)

        Returns
        -------
        list
            Filtered list


        Example:
        --------
        >>> ListExt(range(10)).tail(2)
        [8, 9]
        """
        number_of_items = int(set_min_max(number_of_items, min_value=0, max_value=len(self)))
        return self[::-1][:number_of_items][::-1]

    # Misc
    def apply(self, func: Callable[[T], R]) -> "ListExt2[R]":
        """
        Apply function to each entry

        Parameters
        ----------
        func : Callable[[Any], Any]
            Callable function

        Returns
        -------
        Self
            ListExt


        Example:
        --------
        >>> test = ListExt([1, 2, 3])
        >>> test.apply(str)
        ['1', '2', '3']
        """
        # return self.__class__(map(func, self))
        return self.__class__(func(x) for x in self)  # type: ignore

    @versionchanged("5.2.0", reason="New ``recursive`` parameter")
    def flatten(self, recursive: bool = False) -> "ListExt2":
        """
        Flatten the list

        Parameters
        ----------
        recursive : bool
            Recursively flatten the list, by default ``False``

        Returns
        -------
        Self
            Flattened list


        Example:
        --------
        >>> test = ListExt([["test"], ["test", "test"], ["test"]])
        >>> test.flatten()
        ['test', 'test', 'test', 'test']
        """

        def instance_checking(item):
            if isinstance(item, str):
                return [item]
            if isinstance(item, Iterable):
                return item
            return [item]

        # Flatten
        list_of_list = (instance_checking(x) for x in self)
        flattened = self.__class__(chain(*list_of_list))

        # Recursive
        if recursive:

            def _condition(item) -> bool:
                if isinstance(item, str):
                    return False
                return isinstance(item, Iterable)

            while any(flattened.apply(_condition)):
                flattened = flattened.flatten()

        # Return
        return flattened

    @versionadded("5.2.0")
    def join(self, sep: str = " ", /) -> str:
        """
        Join every element in list to str (``str.join()`` wrapper).

        Parameters
        ----------
        sep : str, optional
            Separator between each element, by default ``" "``

        Returns
        -------
        str
            Joined list


        Example:
        --------
        >>> ListExt(["a", "b", "c"]).join()
        a b c

        >>> # Also work with non-str type
        >>> ListExt([1, 2, 3]).join()
        1 2 3
        """
        try:
            return sep.join(self)  # type: ignore
        except TypeError:
            return sep.join(self.apply(str))

    # Random
    def pick_one(self) -> T:
        """
        Pick one random items from ``list``

        Returns
        -------
        Any
            Random value


        Example:
        --------
        >>> test = ListExt(["foo", "bar"])
        >>> test.pick_one()
        'bar'
        """
        if len(self) != 0:
            out = random.choice(self)
            # logger.debug(out)
            return out
        else:
            # logger.debug("List empty!")
            raise IndexError("List empty!")

    def get_random(self, number_of_items: int = 5, /) -> list[T]:
        """
        Get ``number_of_items`` random items in ``ListExt``

        Parameters
        ----------
        number_of_items : int
            Number random of items, by default ``5``

        Returns
        -------
        list
            Filtered list
        """
        return [self.pick_one() for _ in range(number_of_items)]

    @versionadded("5.10.0")
    def shuffle(self) -> Self:
        """
        Shuffle the list itself

        Returns
        -------
        Self
            Shuffled list
        """
        random.shuffle(self)
        return self

    # Len
    @versionchanged("5.2.0", reason="Handle more type")
    def len_items(self) -> "ListExt2[int]":
        """
        ``len()`` for every item in ``self``

        Returns
        -------
        Self
            List of ``len()``'ed value


        Example:
        --------
        >>> test = ListExt(["foo", "bar", "pizza"])
        >>> test.len_items()
        [3, 3, 5]
        """

        def _len(item: Any) -> int:
            try:
                return len(item)
            except TypeError:
                return len(str(item))

        return self.__class__([_len(x) for x in self])  # type: ignore

    @versionchanged("5.2.0", reason="New ``recursive`` parameter")
    def mean_len(self, recursive: bool = False) -> float:
        """
        Average length of every item. Returns zero if failed (empty list, ZeroDivisionError).

        Parameters
        ----------
        recursive : bool
            Recursively find the average length of items in nested lists, by default ``False``

        Returns
        -------
        float
            Average length


        Example:
        --------
        >>> test = ListExt(["foo", "bar", "pizza"])
        >>> test.mean_len()
        3.6666666666666665
        """

        if recursive:
            dat = self.flatten(recursive=recursive)
        else:
            dat = self

        try:
            return sum(dat.len_items()) / len(dat)
        except ZeroDivisionError:
            return 0.0

    @versionadded("5.2.0")
    def max_item_len(self, recursive: bool = False) -> int:
        """
        Find the maximum length of items in the list.

        Parameters
        ----------
        recursive : bool
            Recursively find the maximum length of items in nested lists, by default ``False``

        Returns
        -------
        int
            Maximum length of items


        Example:
        --------
        >>> test = ListExt(["test", "longer_test"])
        >>> test.max_item_len()
        11

        >>> test = ListExt([["short"], ["longer_test"]])
        >>> test.max_item_len(recursive=True)
        11
        """
        if recursive:
            return cast(int, max(self.flatten(recursive=True).len_items()))
        return cast(int, max(self.len_items()))


if __name__ == "__main__":
    test = ListExt2([1, 2, 3, 4, 5])
    a = test.pick_one()



"""
Absfuyu: Generator
------------------
This generate stuff (Not python's ``generator``)

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)

Features:
---------
- Generate random string
- Generate key
- Generate check digit
- Generate combinations of list in range
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["Charset", "Generator"]


# Library
# ---------------------------------------------------------------------------
import string
from collections.abc import Collection, Sequence
from itertools import chain, combinations
from random import choice
from typing import Literal, TypeVar, cast, overload

from absfuyu.core.baseclass import GetClassMembersMixin
from absfuyu.core.docstring import deprecated
from absfuyu.util import set_min_max

# Type
# ---------------------------------------------------------------------------
T = TypeVar("T")


# Class
# ---------------------------------------------------------------------------
class Charset:
    """
    Character set data class
    """

    DEFAULT: str = string.ascii_letters + string.digits
    ALPHABET: str = string.ascii_letters
    FULL: str = string.ascii_letters + string.digits + string.punctuation
    UPPERCASE: str = string.ascii_uppercase
    LOWERCASE: str = string.ascii_lowercase
    DIGIT: str = string.digits
    SPECIAL: str = string.punctuation
    ALL: str = string.printable
    PRODUCT_KEY: str = (
        "BCDFGHJKMNPQRTVWXY2346789"  # Charset that various key makers use
    )

    def __str__(self) -> str:
        charset = [x for x in self.__class__.__dict__.keys() if not x.startswith("__")]
        return f"List of Character set: {charset}"

    def __repr__(self) -> str:
        charset = [x for x in self.__class__.__dict__.keys() if not x.startswith("__")]
        clsname = self.__class__.__name__
        return f"{clsname}({', '.join(charset)})"


class Generator(GetClassMembersMixin):
    """
    Generator that generate stuffs

    Features:
    ---------
    - Generate random string
    - Generate key
    - Generate check digit
    - Generate combinations of list in range
    """

    # Generate string
    @overload
    @staticmethod
    def generate_string(  # type: ignore
        charset: str = Charset.DEFAULT,
        size: int = 8,
        times: int = 1,
        unique: bool = False,
    ) -> list[str]: ...

    @overload
    @staticmethod
    def generate_string(
        charset: str = Charset.DEFAULT,
        size: int = 8,
        times: int = 1,
        unique: bool = False,
        string_type_if_1: Literal[True] = ...,
    ) -> str: ...

    @staticmethod
    @deprecated("5.2.0", reason="Use generate_string2() instead.")
    def generate_string(
        charset: str = Charset.DEFAULT,
        size: int = 8,
        times: int = 1,
        unique: bool = False,
        string_type_if_1: bool = False,
    ) -> str | list[str]:
        """
        Generate a list of random string from character set (Random string generator).
        Deprecated

        Parameters
        ----------
        charset : str, optional
            - Use custom character set or character sets already defined in ``Charset``
            - ``Charset.DEFAULT``: character in [a-zA-Z0-9] (default)
            - ``Charset.ALPHABET``: character in [a-zA-Z]
            - ``Charset.FULL``: character in [a-zA-Z0-9] + special characters
            - ``Charset.UPPERCASE``: character in [A-Z]
            - ``Charset.LOWERCASE``: character in [a-z]
            - ``Charset.DIGIT``: character in [0-9]
            - ``Charset.SPECIAL``: character in [!@#$%^&*()_+=-]
            - ``Charset.ALL``: character in every printable character

        size : int, optional
            Length of each string in list, by default ``8``

        times : int, optional
            How many random string generated, by default ``1``

        unique : bool, optional
            Each generated text is unique, by default ``False``

        string_type_if_1 : bool, optional
            Return a ``str`` type result if ``times == 1``,
            by default ``False``

        Returns
        -------
        list
            List of random string generated

        str
            When ``string_type_if_1`` is ``True``
        """

        try:
            char_lst = list(charset)
        except Exception:
            char_lst = charset  # type: ignore # ! review this sometime

        unique_string = []
        count = 0

        while count < times:
            s = "".join(choice(char_lst) for _ in range(size))
            if not unique:
                unique_string.append(s)
                count += 1
            else:
                if s not in unique_string:
                    unique_string.append(s)
                    count += 1

        if string_type_if_1 and times == 1:
            return unique_string[0]
        else:
            return unique_string

    @overload
    @staticmethod
    def generate_string2() -> list[str]: ...

    @overload
    @staticmethod
    def generate_string2(
        charset: str = Charset.DEFAULT,
        length: int = 8,
        times: int = 1,
        unique: bool = False,
    ) -> list[str]: ...

    @overload
    @staticmethod
    def generate_string2(
        charset: Sequence[str],
        length: int = 8,
        times: int = 1,
        unique: bool = False,
    ) -> list[str]: ...

    @staticmethod
    def generate_string2(
        charset: Sequence[str] | Charset | str = Charset.DEFAULT,
        length: int = 8,
        times: int = 1,
        unique: bool = False,
    ) -> list[str]:
        """
        Generate a list of random string from character set (Random string generator).
        Improved version.

        Parameters
        ----------
        charset : Sequence[str] | Charset | str, optional
            - Use custom character set or character sets already defined in ``Charset``
            - ``Charset.DEFAULT``: character in [a-zA-Z0-9] (default)
            - ``Charset.ALPHABET``: character in [a-zA-Z]
            - ``Charset.FULL``: character in [a-zA-Z0-9] + special characters
            - ``Charset.UPPERCASE``: character in [A-Z]
            - ``Charset.LOWERCASE``: character in [a-z]
            - ``Charset.DIGIT``: character in [0-9]
            - ``Charset.SPECIAL``: character in [!@#$%^&*()_+=-]
            - ``Charset.ALL``: character in every printable character

        length : int, optional
            Length of each string in list, by default ``8``

        times : int, optional
            How many random string generated, by default ``1``

        unique : bool, optional
            Each generated text is unique, by default ``False``

        Returns
        -------
        list[str]
            List of random generated string.

        Raises
        ------
        TypeError
            When ``charset`` not in correct type.


        Example:
        --------
        >>> Generator.generate_string2(times=3)
        ['67Xfh1fv', 'iChcGz9P', 'u82fNzlm']
        """

        # Type check
        if isinstance(charset, str):
            character_list: list[str] = list(charset)
        elif isinstance(charset, Sequence):
            character_list = cast(list, charset)
        else:
            raise TypeError("Not a valid type.")

        # Setup
        generated_strings: list[str] = []
        unique_generated_strings: set[str] = set[str]()
        times_ = max(1, times)

        while True:
            random_string = "".join(choice(character_list) for _ in range(length))
            if unique:
                unique_generated_strings.add(random_string)
            else:
                generated_strings.append(random_string)

            # Break condition
            if (
                len(unique_generated_strings) >= times_
                or len(generated_strings) >= times_
            ):
                break

        return list(unique_generated_strings) if unique else generated_strings

    # Generate key
    @overload
    @classmethod
    def generate_key(cls) -> str: ...

    @overload
    @classmethod
    def generate_key(
        cls,
        charset: str = Charset.PRODUCT_KEY,
        letter_per_block: int = 5,
        number_of_block: int = 5,
        sep: str = "-",
    ) -> str: ...

    @classmethod
    def generate_key(
        cls,
        charset: str = Charset.PRODUCT_KEY,
        letter_per_block: int = 5,
        number_of_block: int = 5,
        sep: str = "-",
    ) -> str:
        """
        Generate custom key.

        Parameters
        ----------
        charset : Charset | str, optional
            Character set, by default ``Charset.PRODUCT_KEY``

        letter_per_block : int, optional
            Number of letter per key block, by default ``5``

        number_of_block : int, optional
            Number of key block, by default ``5``

        sep : str, optional
            Key block separator, by default ``-``

        Returns
        -------
        str
            Generated key


        Example:
        --------
        >>> Generator.generate_key(letter_per_block=10, number_of_block=2)
        'VKKPJVYD2H-M7R687QCV2'
        """
        out = sep.join(
            cls.generate_string2(
                charset,
                length=letter_per_block,
                times=number_of_block,
                unique=False,
            )
        )
        return out

    # Generate check digit
    @overload
    @staticmethod
    def generate_check_digit(number: int) -> int: ...

    @overload
    @staticmethod
    def generate_check_digit(number: list[int]) -> int: ...

    @overload
    @staticmethod
    def generate_check_digit(number: str) -> int: ...

    @staticmethod
    def generate_check_digit(number: int | list[int] | str) -> int:
        """
        Check digit generator.

        "A check digit is a form of redundancy check used for
        error detection on identification numbers, such as
        bank account numbers, which are used in an application
        where they will at least sometimes be input manually.
        It is analogous to a binary parity bit used to
        check for errors in computer-generated data.
        It consists of one or more digits (or letters) computed
        by an algorithm from the other digits (or letters) in the sequence input.
        With a check digit, one can detect simple errors in
        the input of a series of characters (usually digits)
        such as a single mistyped digit or some permutations
        of two successive digits." (Wikipedia)

        This function use Luhn's algorithm to calculate.

        Parameters
        ----------
        number : int
            Number to calculate check digit

        Returns
        -------
        int
            Check digit


        Example:
        --------
        >>> Generator.generate_check_digit(4129984562545)
        7
        """

        # Type handle
        if isinstance(number, int):
            # Turn into list[int] then reverse the order
            num = [int(x) for x in list(str(number))][::-1]
        elif isinstance(number, str):
            num = [int(x) for x in list(number)][::-1]
        elif isinstance(number, list):
            num = number[::-1]
        else:
            raise TypeError("Variable `number` is not correct type.")

        sum = 0

        # for i in range(len(num)):
        for idx, _ in enumerate(num):
            if idx % 2 == 0:
                # double value of the even-th digit
                num[idx] *= 2
                # sum the character of digit if it's >= 10
                if num[idx] >= 10:
                    num[idx] -= 9
            sum += num[idx]

        out = (10 - (sum % 10)) % 10
        return out

    # Generate combinations range
    @overload
    @staticmethod
    def combinations_range(collection: Collection[T]) -> list[tuple[T, ...]]: ...

    @overload
    @staticmethod
    def combinations_range(collection: Collection[T], *, min_len: int = 1, max_len: int = 0) -> list[tuple[T, ...]]: ...

    @staticmethod
    def combinations_range(
        collection: Collection[T], *, min_len: int = 1, max_len: int = 0
    ) -> list[tuple[T, ...]]:
        """
        Generate all combinations of a ``collection`` from ``min_len`` to ``max_len``

        Parameters
        ----------
        collection : Collection[T]
            A collection (Iterable with ``__len__``) that need to generate combination

        min_len : int, optional
            Minimum ``r`` of ``combinations``, by default ``1``

        max_len : int, optional
            Maximum ``r`` of ``combinations``, by default ``0`` (len of ``collection``)

        Returns
        -------
        list[tuple]
            A list of all combinations from range(``min_len``, ``max_len``) of ``collection``


        Example:
        --------
        >>> Generator.combinations_range([1, 2, 3], min_len=2)
        [(1, 2), (1, 3), (2, 3), (1, 2, 3)]
        """
        # Restrain
        len_iter = len(
            collection
        )  # with collection.abc.Iterable, use len(list(collection))
        if max_len < 1:
            max_len = len_iter
        max_len = int(min(max_len, len_iter))
        min_len = int(set_min_max(min_len, min_value=1, max_value=max_len))

        # Return
        return list(
            chain.from_iterable(
                combinations(collection, i) for i in range(min_len, max_len + 1)
            )
        )



"""
Test: Util - API

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

import sys

import pytest

try:
    import requests  # type: ignore
except ImportError:
    requests = pytest.importorskip("requests")

from absfuyu.util.api import APIRequest, PingResult, ping_windows


@pytest.mark.abs_util
class TestUtilAPI:
    """absfuyu.util.api"""

    def test_API(self, mocker) -> None:
        mock_APIRequest = mocker.Mock(spec=APIRequest)  # Mock APIRequest
        mock_APIRequest.fetch_data_only.return_value.json.return_value = {
            "quotes": ["a", "b"]
        }

        instance: APIRequest = mock_APIRequest
        assert isinstance(instance.fetch_data_only().json()["quotes"], list)

    @pytest.mark.skipif(
        sys.platform not in ["win32", "cygwin"],
        reason="Not applicable on Linux and MacOS",
    )  # windows only
    def test_ping_windows(self, mocker) -> None:
        mock_ping = mocker.patch("absfuyu.util.api.ping_windows")
        mock_ping.return_value = [PingResult("", "")]
        res = mock_ping(["google.com"], 1)
        assert isinstance(res[0], PingResult)



"""
Test: Logger

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Library
# ---------------------------------------------------------------------------
import pytest

from absfuyu.logger import compress_for_log


# Test
# ---------------------------------------------------------------------------
def test_list():
    test_list = ["test"] * 200
    assert compress_for_log(test_list)


def test_set():
    test_set = set(range(200))
    assert compress_for_log(test_set)


def test_tuple():
    test_tuple = tuple(["test"] * 200)
    assert compress_for_log(test_tuple)


def test_dict():
    test_dict = dict(zip(range(200), range(200)))
    assert compress_for_log(test_dict)


def test_str():
    test_string = "test\n\n" * 200
    assert compress_for_log(test_string)


def test_object():
    import absfuyu

    assert compress_for_log(absfuyu)



"""
Absfuyu: Core
-------------
Bases for other features

Version: 5.15.0
Date updated: 03/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    # Color
    "CLITextColor",
    # Support
    "ClassMembers",
    "ClassMembersResult",
    # Mixins
    "GetClassMembersMixin",
    "AutoREPRMixin",
    # Class
    "BaseClass",
    "BaseDataclass",
    # Metaclass
    "PositiveInitArgsMeta",
]

# Library
# ---------------------------------------------------------------------------
from dataclasses import dataclass, field
from typing import Any, ClassVar, Literal, Self


# Color
# ---------------------------------------------------------------------------
class CLITextColor:
    """Color code for text in terminal"""

    WHITE = "\x1b[37m"
    BLACK = "\x1b[30m"
    BLUE = "\x1b[34m"
    GRAY = "\x1b[90m"
    GREEN = "\x1b[32m"
    RED = "\x1b[91m"
    DARK_RED = "\x1b[31m"
    MAGENTA = "\x1b[35m"
    YELLOW = "\x1b[33m"
    RESET = "\x1b[39m"


# Dataclass
# ---------------------------------------------------------------------------
# @versionadded("5.5.0")
@dataclass
class BaseDataclass:
    """
    Base dataclass.

    Contains util methods:
    - _get_fields
    - to_dict
    """

    @classmethod
    def _get_fields(cls) -> tuple[str, ...]:
        """
        Get dataclass's fields.

        Returns
        -------
        tuple[str, ...]
            Dataclass's fields.
        """
        _fields = getattr(cls, "__dataclass_fields__", ())
        return tuple(_fields)

    # @versionadded("5.11.0")
    def to_dict(self):
        """
        Convert dataclass into dict

        Returns
        -------
        dict
            Dataclass converted into dict
        """
        fields = self._get_fields()
        output = {}
        for field in fields:
            output.__setitem__(field, self.__getattribute__(field))
        return output


# Support
# ---------------------------------------------------------------------------
# @versionadded("5.5.0")
@dataclass
class ClassMembers(BaseDataclass):
    """
    Contains lists of methods, classmethods, staticmethods,
    properties, and attributes of a class.

    Parameters
    ----------
    methods : list[str], optional
        List contains method names of a class.
        By default inits an empty list.

    classmethods : list[str], optional
        List contains classmethod names of a class.
        By default inits an empty list.

    staticmethods : list[str], optional
        List contains staticmethod names of a class.
        By default inits an empty list.

    properties : list[str], optional
        List contains property names of a class.
        By default inits an empty list.

    attributes : list[str], optional
        List contains attributes of a class instance.
        By default inits an empty list.

    classattributes : list[str], optional
        List contains attributes of a class.
        By default inits an empty list.


    Available formats
    -----------------
    Alternative format to use with: ``format(<obj>, <format_spec>)``
    - ``s``/``short`` (This hides attributes with empty value)
    """

    methods: list[str] = field(default_factory=list)
    classmethods: list[str] = field(default_factory=list)
    staticmethods: list[str] = field(default_factory=list)
    properties: list[str] = field(default_factory=list)
    attributes: list[str] = field(default_factory=list)
    classattributes: list[str] = field(default_factory=list)

    def __format__(self, format_spec: str) -> str:
        """
        Available format:
        - ``s``/``short`` (This hides attributes with empty value)

        Parameters
        ----------
        format_spec : str
            Format spec

        Returns
        -------
        str
            Alternative text representation of this class.
        """

        fmt = format_spec.lower().strip()
        cls_name = self.__class__.__name__

        if fmt == "short" or fmt == "s":
            out = []
            sep = ", "
            for x in self._get_fields():
                if len(getattr(self, x)) > 0:
                    out.append(f"{x}={repr(getattr(self, x))}")
            return f"{cls_name}({sep.join(out)})"

        return self.__repr__()

    def is_empty(self) -> bool:
        """
        Checks if all fields are empty.


        Example:
        --------
        >>> ClassMembers().is_empty()
        True

        >>> ClassMembers(["a"]).is_empty()
        False
        """

        # return all(len(getattr(self, x)) == 0 for x in self._get_fields())
        for x in self._get_fields():
            if len(getattr(self, x)) > 0:
                return False
        return True

    def pack(
        self,
        include_method: bool = True,
        include_classmethod: bool = True,
        classmethod_indicator: str = "<cls>",
        include_staticmethod: bool = True,
        staticmethod_indicator: str = "<stc>",
        include_attribute: bool = True,
        include_classattribute: bool = True,
        classattribute_indicator: str = "<cls>",
    ) -> Self:
        """
        Combines the following into one list:
        - methods, classmethods, and staticmethods
        - attributes and class attributes

        Parameters
        ----------
        include_method : bool, optional
            Whether to include methods in the output, by default ``True``

        include_classmethod : bool, optional
            Whether to include classmethods in the output, by default ``True``

        classmethod_indicator : str, optional
            A string used to mark classmethod in the output. This string is appended
            to the name of each classmethod to visually differentiate it from regular
            instance methods, by default ``"<cls>"``

        include_staticmethod : bool, optional
            Whether to include staticmethods in the output, by default ``True``

        staticmethod_indicator : str, optional
            A string used to mark staticmethod in the output. This string is appended
            to the name of each staticmethod to visually differentiate it from regular
            instance methods, by default ``"<stc>"``

        include_attribute : bool, optional
            Whether to include attributes in the output, by default ``True``

        include_classattribute : bool, optional
            Whether to include class attributes in the output, by default ``True``

        classattribute_indicator : str, optional
            A string used to mark class attribute in the output. This string is appended
            to the name of each class attribute to visually differentiate it from regular
            instance attributes, by default ``"<cls>"``

        Returns
        -------
        Self
            ClassMembers (combined methods)


        Example:
        --------
        >>> ClassMembers(["a"], ["b"], ["c"], ["d"], ["e"], ["f"]).pack().__format__("short")
        ClassMembers(methods=['a', 'b <cls>', 'c <stc>'], properties=['d'], attributes=['e', 'f <cls>'])
        """

        new_methods_list: list[str] = []

        # Method
        if include_method:
            new_methods_list.extend(self.methods)

        # Classmethod
        if include_classmethod:
            new_methods_list.extend(
                [f"{x} {classmethod_indicator}".strip() for x in self.classmethods]
            )

        # Staticmethod
        if include_staticmethod:
            new_methods_list.extend(
                [f"{x} {staticmethod_indicator}".strip() for x in self.staticmethods]
            )

        new_attributes_list: list[str] = []

        # Attribute
        if include_attribute:
            new_attributes_list.extend(self.attributes)

        # Class attribute
        if include_classattribute:
            new_attributes_list.extend(
                [
                    f"{x} {classattribute_indicator}".strip()
                    for x in self.classattributes
                ]
            )

        # return self.__class__(new_methods_list, [], [], self.properties, [])
        return self.__class__(
            methods=new_methods_list,
            properties=self.properties,
            attributes=new_attributes_list,
        )

    def sort(self, reverse: bool = False) -> Self:
        """
        Sorts every element in each method list.

        Parameters
        ----------
        reverse : bool, optional
            Descending order, by default ``False``

        Returns
        -------
        Self
            Self with sorted values.


        Example:
        --------
        >>> ClassMembers(["b", "a"], ["d", "c"], ["f", "e"]).sort().__format__("short")
        ClassMembers(methods=['a', 'b'], classmethods=['c', 'd'], staticmethods=['e', 'f'])

        >>> ClassMembers(["b", "a"], ["d", "c"], ["f", "e"]).pack().sort().__format__("short")
        ClassMembers(methods=['a', 'b', 'c <cls>', 'd <cls>', 'e <stc>', 'f <stc>'])
        """

        sorted_vals: list[list[str]] = [
            sorted(getattr(self, field), reverse=reverse)
            for field in self._get_fields()
        ]

        return self.__class__(*sorted_vals)


# @versionadded("5.5.0")
class ClassMembersResult(dict[str, ClassMembers]):
    """
    All members of a class and its parent classes.
    """

    _LINELENGTH: ClassVar[int] = 88

    def __format__(self, format_spec: str) -> str:
        """
        Available format:
        - ``s``/``short`` (This hides attributes with empty value)

        Parameters
        ----------
        format_spec : str
            Format spec

        Returns
        -------
        str
            Alternative text representation of this class.
        """

        fmt = format_spec.lower().strip()

        if fmt == "short" or fmt == "s":
            # out = {}
            # for name, member in self.items():
            #     out[name] = format(member, fmt)
            # return repr(out)
            out = []
            sep = ", "
            for idx, (name, member) in enumerate(self.items()):
                mem = format(member, fmt)
                base = f"{repr(name)}: {mem}"

                # if idx == 0:
                #     out.append("{" + base)
                # elif idx + 1 == len(self.items()):
                #     out.append(base + "}")
                # else:
                #     out.append(base)
                out.append(
                    ("{" if idx == 0 else "")
                    + base
                    + ("}" if idx + 1 == len(self.items()) else "")
                )

            return sep.join(out)

        return self.__repr__()

    def _merge_value(
        self,
        value_name: Literal[
            "methods",
            "classmethods",
            "staticmethods",
            "properties",
            "attributes",
            "classattributes",
        ],
    ) -> list[str]:
        """
        Merge all specified values from the dictionary.

        Parameters
        ----------
        value_name : Literal["methods", "classmethods", "staticmethods", "properties", "attributes", "classattributes"]
            The type of value to merge.

        Returns
        -------
        list[str]
            A list of merged values.
        """

        merged = []
        for _, member in self.items():
            if value_name in member._get_fields():
                merged.extend(getattr(member, value_name))
        return merged

    def flatten_value(self, sort: bool = True) -> ClassMembers:
        """
        Merge all attributes of ``dict``'s values into one ``ClassMembers``.

        Parameters
        ----------
        sort : bool
            Sort value in ascending order after flatten, by default ``True``

        Returns
        -------
        ClassMembers
            Flattened value.


        Example:
        --------
        >>> test = ClassMembersResult(
        ...     ABC=ClassMembers(["a"], ["b"], ["c"], ["d"], ["y"], ["x"]),
        ...     DEF=ClassMembers(["e"], ["f"], ["g"], ["h"], ["w"], ["z"]),
        ... )
        >>> test.flatten_value()
        ClassMembers(
            methods=["a", "e"],
            classmethods=["b", "f"],
            staticmethods=["c", "g"],
            properties=["d", "h"],
            attributes=["w", "y"],
            classattributes=["x", "z"],
        )
        """

        res: list[list[str]] = []
        # for x in [
        #     "methods",
        #     "classmethods",
        #     "staticmethods",
        #     "properties",
        #     "attributes",
        #     "classattributes",
        # ]:
        for x in ClassMembers._get_fields():
            res.append(self._merge_value(x))  # type: ignore
        return ClassMembers(*res).sort() if sort else ClassMembers(*res)

    def pack_value(
        self,
        include_method: bool = True,
        include_classmethod: bool = True,
        classmethod_indicator: str = "<classmethod>",
        include_staticmethod: bool = True,
        staticmethod_indicator: str = "<staticmethod>",
        include_attribute: bool = True,
        include_classattribute: bool = True,
        classattribute_indicator: str = "<cls>",
    ) -> Self:
        """
        Join members into one list for each value.

        Parameters
        ----------
        include_method : bool, optional
            Whether to include method in the output, by default ``True``

        include_classmethod : bool, optional
            Whether to include classmethod in the output, by default ``True``

        classmethod_indicator : str, optional
            A string used to mark classmethod in the output. This string is appended
            to the name of each classmethod to visually differentiate it from regular
            instance methods, by default ``"<classmethod>"``

        include_staticmethod : bool, optional
            Whether to include staticmethod in the output, by default ``True``

        staticmethod_indicator : str, optional
            A string used to mark staticmethod in the output. This string is appended
            to the name of each staticmethod to visually differentiate it from regular
            instance methods, by default ``"<staticmethod>"``

        include_attribute : bool, optional
            Whether to include attributes in the output, by default ``True``

        include_classattribute : bool, optional
            Whether to include class attributes in the output, by default ``True``

        classattribute_indicator : str, optional
            A string used to mark class attribute in the output. This string is appended
            to the name of each class attribute to visually differentiate it from regular
            instance attributes, by default ``"<cls>"``

        Returns
        -------
        Self
            ClassMembersResult with packed value.


        Example:
        --------
        >>> test = ClassMembersResult(
        ...     ABC=ClassMembers(["a"], ["b"], ["c"], ["d"], ["y"], ["x"]),
        ...     DEF=ClassMembers(["e"], ["f"], ["g"], ["h"], ["w"], ["z"]),
        ... )
        >>> test.pack_value()
        {
            "ABC": ClassMembers(
                methods=["a", "b <classmethod>", "c <staticmethod>"],
                classmethods=[],
                staticmethods=[],
                properties=["d"],
                attributes=["y", "x <cls>"],
                classattributes=[],
            ),
            "DEF": ClassMembers(
                methods=["e", "f <classmethod>", "g <staticmethod>"],
                classmethods=[],
                staticmethods=[],
                properties=["h"],
                attributes=["w", "z <cls>"],
                classattributes=[],
            ),
        }
        """

        for class_name, members in self.items():
            self[class_name] = members.pack(
                include_method=include_method,
                include_classmethod=include_classmethod,
                classmethod_indicator=classmethod_indicator,
                include_staticmethod=include_staticmethod,
                staticmethod_indicator=staticmethod_indicator,
                include_attribute=include_attribute,
                include_classattribute=include_classattribute,
                classattribute_indicator=classattribute_indicator,
            )
        return self

    def prioritize_value(
        self,
        value_name: Literal[
            "methods",
            "classmethods",
            "staticmethods",
            "properties",
            "attributes",
            "classattributes",
        ] = "methods",
    ) -> dict[str, list[str]]:
        """
        Prioritize which field of value to show.

        Parameters
        ----------
        value_name : Literal["methods", "classmethods", "staticmethods", "properties", "attributes", "classattributes"], optional
            The type of value to prioritize, by default ``"methods"``

        Returns
        -------
        dict[str, list[str]]
            A dictionary with prioritized values.


        Example:
        --------
        >>> test = ClassMembersResult(
        ...     ABC=ClassMembers(["a"], ["b"], ["c"], ["d"]),
        ...     DEF=ClassMembers(["e"], ["f"], ["g"], ["h"]),
        ... )
        >>> test.prioritize_value("methods")
        {'ABC': ['a'], 'DEF': ['e']}
        >>> test.prioritize_value("classmethods")
        {'ABC': ['b'], 'DEF': ['f']}
        >>> test.prioritize_value("staticmethods")
        {'ABC': ['c'], 'DEF': ['g']}
        >>> test.prioritize_value("properties")
        {'ABC': ['d'], 'DEF': ['h']}
        """

        result: dict[str, list[str]] = {}
        for name, member in self.items():
            result[name] = getattr(member, value_name, member.methods)
        return result

    def print_output(
        self,
        where_to_print: Literal["methods", "properties", "attributes"] = "methods",
        print_in_one_column: bool = False,
    ) -> None:
        """
        Beautifully print the result.

        *This method is deprecated.*

        Parameters
        ----------
        where_to_print : Literal["methods", "properties", "attributes"], optional
            Whether to print ``methods`` or ``properties``, by default ``"methods"``

        print_in_one_column : bool, optional
            Whether to print in one column, by default ``False``
        """

        print_func = print  # Can be extended with function parameter

        # Loop through each class base
        for order, (class_base, member) in enumerate(self.items(), start=1):
            methods: list[str] = getattr(member, where_to_print, member.methods)
            mlen = len(methods)  # How many methods in that class
            if mlen == 0:
                continue
            print_func(f"{order:02}. <{class_base}> | len: {mlen:02}")

            # Modify methods list
            max_method_name_len = max([len(x) for x in methods])
            if mlen % 2 == 0:
                p1, p2 = methods[: int(mlen / 2)], methods[int(mlen / 2) :]
            else:
                p1, p2 = methods[: int(mlen / 2) + 1], methods[int(mlen / 2) + 1 :]
                p2.append("")
            new_methods = list(zip(p1, p2))

            # print
            if print_in_one_column:
                # This print 1 method in one line
                for name in methods:
                    print(f"    - {name.ljust(max_method_name_len)}")
            else:
                # This print 2 methods in 1 line
                for x1, x2 in new_methods:
                    if x2 == "":
                        print_func(f"    - {x1.ljust(max_method_name_len)}")
                    else:
                        print_func(
                            f"    - {x1.ljust(max_method_name_len)}    - {x2.ljust(max_method_name_len)}"
                        )

            print_func("".ljust(self._LINELENGTH, "-"))


# Mixins
# ---------------------------------------------------------------------------
class GetClassMembersMixin:
    """
    Show all methods of the class and its parent class minus ``object`` class

    *This class is meant to be used with other class*


    Example:
    --------
    >>> class TestClass(GetClassMembersMixin):
    ...     def method1(self): ...
    >>> TestClass._get_members(dunder=False)
    {
        'GetClassMembersMixin': ClassMembers(
            methods=[],
            classmethods=['_get_members', ..., 'show_all_methods', 'show_all_properties'],
            staticmethods=[],
            properties=[],
            attributes=[],
            classattributes=[]
        ),
        'TestClass': ClassMembers(
            methods=['method1'],
            classmethods=[],
            staticmethods=[],
            properties=[],
            attributes=[],
            classattributes=[]
        )
    }
    """

    # @versionadded("5.5.0")
    @classmethod
    def _get_members(
        cls,
        dunder: bool = True,
        underscore: bool = True,
        private: bool = True,
    ) -> ClassMembersResult:
        """
        Class method to get all methods, properties, and attributes
        of the class and its parent classes.

        Parameters
        ----------
        dunder : bool, optional
            Whether to include attribute with ``__`` (dunder)
            in the output, by default ``True``

        underscore : bool, optional
            Whether to include attribute starts with ``_``
            in the output (will also skip dunder), by default ``True``

        private : bool, optional
            Whether to include private attribute
            (``_<__class__.__name__>__<attribute>``)
            in the output, by default ``True``

        Returns
        -------
        ClassMembersResult
            A dictionary where keys are class names
            and values are tuples of method names and properties.
        """

        # MRO in reverse order
        classes = cls.__mro__[::-1]
        result: dict[str, ClassMembers] = {}

        # For each class base in classes
        for base in classes:
            methods: list[str] = []
            classmethods: list[str] = []
            staticmethods: list[str] = []
            properties: list[str] = []
            classattributes: list[str] = []

            # Dict items of base
            for name, attr in base.__dict__.items():
                # Skip dunder
                if name.startswith("__") and not dunder:
                    continue

                # Skip private attribute
                if base.__name__ in name and not private:
                    continue

                # Skip underscore
                if (
                    name.startswith("_")
                    and not underscore
                    and base.__name__ not in name
                ):
                    continue

                # Methods
                if callable(attr):
                    if isinstance(attr, staticmethod):
                        staticmethods.append(name)
                    else:
                        methods.append(name)
                elif isinstance(attr, classmethod):
                    classmethods.append(name)

                # Property
                elif isinstance(attr, property):
                    properties.append(name)

                # Class attribute
                else:
                    classattributes.append(name)

                # Save to result
                result[base.__name__] = ClassMembers(
                    methods=methods,
                    classmethods=classmethods,
                    staticmethods=staticmethods,
                    properties=properties,
                    classattributes=classattributes,
                ).sort()

        return ClassMembersResult(result)

    # @versionadded("5.5.0")
    def _get_attributes(
        self,
        underscore: bool = True,
        private: bool = True,
    ) -> list[str]:
        """
        Get all attributes of the class instance.

        Parameters
        ----------
        underscore : bool, optional
            Whether to include attribute starts with ``_``
            in the output, by default ``True``

        private : bool, optional
            Whether to include private attribute
            (``_<__class__.__name__>__<attribute>``)
            in the output, by default ``True``

        Returns
        -------
        list[str]
            A list contains attributes.
        """

        # Default output
        out: list[str] = []

        # Get attributes
        cls_dict: dict[str, Any] | None = getattr(self, "__dict__", None)
        cls_slots: tuple[str, ...] | None = getattr(self, "__slots__", None)
        cls_name: str = self.__class__.__name__

        def _is_valid(item_name: str) -> bool:
            if not item_name.startswith("_"):
                return True
            if cls_name in item_name and private:
                return True
            if underscore and cls_name not in item_name:
                return True
            return False

        # Check if __dict__ exist and len(__dict__) > 0
        if cls_dict is not None and len(cls_dict) > 0:
            # out = [x for x in self.__dict__ if _is_valid(x)]
            out.extend(list(cls_dict))

        # Check if __slots__ exist and len(__slots__) > 0
        if cls_slots is not None and len(cls_slots) > 0:
            # Convert __<attribute> to _<self.__class__.__name__>__<attribute>
            _slot = [f"_{cls_name}{x}" if x.startswith("__") else x for x in cls_slots]
            out.extend(_slot)

        return [x for x in out if _is_valid(x)]

    # @versionadded("5.5.0")
    def get_members(
        self,
        dunder: bool = False,
        underscore: bool = True,
        private: bool = True,
    ) -> ClassMembersResult:
        """
        Get all members of a class instance.

        Parameters
        ----------
        dunder : bool, optional
            Whether to include attribute with ``__`` (dunder)
            in the output, by default ``False``

        underscore : bool, optional
            Whether to include attribute starts with ``_``
            in the output (will also skip dunder), by default ``True``

        private : bool, optional
            Whether to include private attribute
            (``_<__class__.__name__>__<attribute>``)
            in the output, by default ``True``

        Returns
        -------
        ClassMembersResult
            All member of a class instance.
        """
        mems = self._get_members(dunder=dunder, underscore=underscore, private=private)
        attrs = self._get_attributes(underscore=underscore, private=private)
        mems[self.__class__.__name__].attributes = attrs
        return mems

    # @remove("6.0.0")
    # @deprecated("5.5.0")
    # @versionadded("5.1.0")
    @classmethod
    def _get_methods_and_properties(
        cls,
        skip_private_attribute: bool = True,
        include_private_method: bool = False,
    ) -> ClassMembersResult:
        """Wrapper of ``cls._get_members``, deprecated"""
        return cls._get_members(
            dunder=skip_private_attribute, private=include_private_method
        )

    @classmethod
    def show_all_methods(
        cls,
        print_result: bool = False,
        include_classmethod: bool = True,
        classmethod_indicator: str = "<classmethod>",
        include_staticmethod: bool = True,
        staticmethod_indicator: str = "<staticmethod>",
        include_private_method: bool = False,
    ) -> dict[str, list[str]]:
        """
        Class method to display all methods of the class and its parent classes,
        including the class in which they are defined in alphabetical order.

        Parameters
        ----------
        print_result : bool, optional
            Beautifully print the output, by default ``False``

        include_classmethod : bool, optional
            Whether to include classmethod in the output, by default ``True``

        classmethod_indicator : str, optional
            A string used to mark classmethod in the output. This string is appended
            to the name of each classmethod to visually differentiate it from regular
            instance methods, by default ``"<classmethod>"``

        include_staticmethod : bool, optional
            Whether to include staticmethod in the output, by default ``True``

        staticmethod_indicator : str, optional
            A string used to mark staticmethod in the output. This string is appended
            to the name of each staticmethod to visually differentiate it from regular
            instance methods, by default ``"<staticmethod>"``

        include_private_method : bool, optional
            Whether to include private method in the output, by default ``False``

        Returns
        -------
        dict[str, list[str]]
            A dictionary where keys are class names and values are lists of method names.
        """

        result = cls._get_members(
            dunder=False, private=include_private_method
        ).pack_value(
            include_classmethod=include_classmethod,
            classmethod_indicator=classmethod_indicator,
            include_staticmethod=include_staticmethod,
            staticmethod_indicator=staticmethod_indicator,
        )

        if print_result:
            result.print_output("methods")

        return result.prioritize_value("methods")

    @classmethod
    def show_all_properties(cls, print_result: bool = False) -> dict[str, list[str]]:
        """
        Class method to display all properties of the class and its parent classes,
        including the class in which they are defined in alphabetical order.

        Parameters
        ----------
        print_result : bool, optional
            Beautifully print the output, by default ``False``

        Returns
        -------
        dict[str, list[str]]
            A dictionary where keys are class names and values are lists of property names.
        """

        # result = cls.get_methods_and_properties().prioritize_value("properties")
        result = ClassMembersResult(
            {
                cls.__name__: ClassMembers(
                    properties=cls._get_members(dunder=False)
                    .flatten_value()
                    .properties,
                )
            }
        )

        if print_result:
            result.print_output("properties")

        return result.prioritize_value("properties")


# Temp name for backward compability
# Will be removed at version 6.0.0
MethodNPropertyList = ClassMembers
MethodNPropertyResult = ClassMembersResult
ShowAllMethodsMixin = GetClassMembersMixin


class AutoREPRMixin:
    """
    Generate ``repr()`` output as ``<class(param1=any, param2=any, ...)>``

    *This class is meant to be used with other class*


    Example:
    --------
    >>> class Test(AutoREPRMixin):
    ...     def __init__(self, param):
    ...         self.param = param
    >>> print(repr(Test(1)))
    Test(param=1)
    """

    def __repr__(self) -> str:
        """
        Generate a string representation of the instance's attributes.

        This function retrieves attributes from either the ``__dict__`` or
        ``__slots__`` of the instance, excluding private attributes (those
        starting with an underscore). The attributes are returned as a
        formatted string, with each attribute represented as ``"key=value"``.

        Convert ``self.__dict__`` from ``{"a": "b"}`` to ``a=repr(b)``
        or ``self.__slots__`` from ``("a",)`` to ``a=repr(self.a)``
        (excluding private attributes)
        """
        # Default output
        out = []
        sep = ", "  # Separator

        # Get attributes
        cls_dict = getattr(self, "__dict__", None)
        cls_slots = getattr(self, "__slots__", None)

        # Check if __dict__ exist and len(__dict__) > 0
        if cls_dict is not None and len(cls_dict) > 0:
            out = [
                f"{k}={repr(v)}"
                for k, v in self.__dict__.items()
                if not k.startswith("_")
            ]

        # Check if __slots__ exist and len(__slots__) > 0
        elif cls_slots is not None and len(cls_slots) > 0:
            out = [
                f"{x}={repr(getattr(self, x))}"
                for x in self.__slots__  # type: ignore
                if not x.startswith("_")
            ]

        # Return out
        return f"{self.__class__.__name__}({sep.join(out)})"


# Class
# ---------------------------------------------------------------------------
class BaseClass(GetClassMembersMixin, AutoREPRMixin):
    """Base class"""

    def __str__(self) -> str:
        return repr(self)

    def __format__(self, format_spec: str) -> str:
        """
        Formats the object according to the specified format.
        If no format_spec is provided, returns the object's string representation.
        (Currently a dummy function)

        Usage
        -----
        >>> print(f"{<object>:<format_spec>}")
        >>> print(<object>.__format__(<format_spec>))
        >>> print(format(<object>, <format_spec>))
        """

        return self.__str__()


# Metaclass
# ---------------------------------------------------------------------------
class PositiveInitArgsMeta(type):
    """
    Make sure that every args in a class __init__ is positive

    Usage
    -----
    >>> class Test(metaclass=PositiveInitArgsMeta): pass
    """

    def __call__(cls, *args, **kwargs):
        # Check if all positional and keyword arguments are positive
        for arg in args:
            if isinstance(arg, (int, float)) and arg < 0:
                raise ValueError(f"Argument {arg} must be positive")
        for key, value in kwargs.items():
            if isinstance(value, (int, float)) and value < 0:
                raise ValueError(f"Argument {key}={value} must be positive")

        # Call the original __init__ method
        return super().__call__(*args, **kwargs)
